#!/usr/bin/env python3
#
# Tests the lib.guess variable meaning guessing module
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.lib.guess as guess


class LibGuessTest(unittest.TestCase):
    """
    Tests for ``myokit.lib.guess``.
    """

    def test_add_embedded_protocol_pacing(self):
        # Tests the method to add protocols, to models without a stimulus
        # current

        # 1. Should work for variable bound to pace, and with protocol
        m = myokit.Model('m')
        c = m.add_component('c')
        t = c.add_variable('t')
        t.set_unit(myokit.units.second)
        t.set_binding('time')
        x = c.add_variable('x')
        xa = x.add_variable('a')
        xa.set_rhs(3)
        xb = x.add_variable('b')
        xb.set_rhs('3 * a')
        x.set_rhs('sqrt(b) * 3 [mV]')
        x.set_unit(myokit.units.ampere)
        x.set_binding('pace')

        protocol = myokit.pacing.blocktrain(
            duration=1e-3, offset=0.01, period=1, level=2)

        # Check adding works
        model = m.clone()
        self.assertTrue(guess.add_embedded_protocol(model, protocol))

        # Check generated rhs
        self.assertTrue(isinstance(model.get('c.x').rhs(), myokit.If))
        self.assertEqual(
            model.get('c.x').rhs().code(),
            'if((c.t - offset) % period < duration, 2 [A], 0 [A])')
        self.assertIsNone(model.get('c.x').binding())

        # Check units
        self.assertEqual(
            model.get('c.x').unit(), myokit.units.ampere)
        self.assertEqual(
            model.get('c.x.offset').unit(), myokit.units.second)
        self.assertEqual(
            model.get('c.x.duration').unit(), myokit.units.second)
        self.assertEqual(
            model.get('c.x.period').unit(), myokit.units.second)

        # Check values
        self.assertEqual(model.get('c.x.offset').rhs(),
                         myokit.Number(0.01, myokit.units.second))
        self.assertEqual(model.get('c.x.duration').rhs(),
                         myokit.Number(1e-3, myokit.units.second))
        self.assertEqual(model.get('c.x.period').rhs(),
                         myokit.Number(1, myokit.units.second))

        # 2. Should fail if nothing bound to pace
        x.set_binding(None)
        code = m.code()
        self.assertFalse(guess.add_embedded_protocol(m, protocol))
        self.assertEqual(code, m.code())
        x.set_binding('pace')
        model = m.clone()
        self.assertTrue(guess.add_embedded_protocol(model, protocol))
        self.assertTrue(isinstance(model.get('c.x').rhs(), myokit.If))

        # 3 .Should fail if there's no time variable
        t.set_binding(None)
        code = m.code()
        self.assertFalse(guess.add_embedded_protocol(m, protocol))
        self.assertEqual(code, m.code())
        t.set_binding('time')
        model = m.clone()
        self.assertTrue(guess.add_embedded_protocol(model, protocol))
        self.assertTrue(isinstance(model.get('c.x').rhs(), myokit.If))

        # 4. Should fail if we have a step protocol
        p = myokit.pacing.steptrain([1, 2, 3], -80, 0.01, 1)
        code = m.code()
        self.assertFalse(guess.add_embedded_protocol(m, p))
        self.assertEqual(code, m.code())
        model = m.clone()
        self.assertTrue(guess.add_embedded_protocol(model, protocol))
        self.assertTrue(isinstance(model.get('c.x').rhs(), myokit.If))

        # 5. Should fail if we have a non-zero multiplier
        protocol = myokit.Protocol()
        protocol.schedule(
            duration=1e-3, start=0.01, period=1, level=2, multiplier=10)
        code = m.code()
        self.assertFalse(guess.add_embedded_protocol(m, protocol))
        self.assertEqual(code, m.code())

    def test_add_embedded_protocol_stimulus(self):
        # Tests the method to add protocols to models with a stimulus current

        # 1. Should work for variable bound to pace, and with protocol
        m = myokit.parse_model("""
        [[model]]

        [engine]
        time = 0 [ms]
            in [ms]
            bind time
        pace = x
            x = 5 [1]
            bind pace
            in [1]

        [stimulus]
        i_stim = engine.pace * amplitude
            in [pA]
        amplitude = -80 [pA]
            in [pA]
        """)

        protocol = myokit.pacing.blocktrain(
            duration=0.5, offset=5, period=1000, level=2)

        # Check adding works
        model = m.clone()
        self.assertTrue(guess.add_embedded_protocol(model, protocol))

        # Check generated rhs
        rhs = model.get('stimulus.i_stim').rhs()
        self.assertTrue(isinstance(rhs, myokit.Multiply))
        self.assertTrue(isinstance(rhs[0], myokit.If))

        # Check pacing variable has disappeared
        self.assertNotIn('pace', model.get('engine'))
        self.assertIsNone(model.binding('pace'))

        # Check units
        self.assertEqual(
            model.get('stimulus.i_stim').unit(),
            myokit.units.pA)
        self.assertEqual(
            model.get('stimulus.i_stim.offset').unit(),
            myokit.units.ms)
        self.assertEqual(
            model.get('stimulus.i_stim.duration').unit(),
            myokit.units.ms)
        self.assertEqual(
            model.get('stimulus.i_stim.period').unit(),
            myokit.units.ms)

        # Check values
        self.assertEqual(model.get('stimulus.i_stim.offset').rhs(),
                         myokit.Number(5, myokit.units.ms))
        self.assertEqual(model.get('stimulus.i_stim.duration').rhs(),
                         myokit.Number(0.5, myokit.units.ms))
        self.assertEqual(model.get('stimulus.i_stim.period').rhs(),
                         myokit.Number(1000, myokit.units.ms))

        # 2. Should fail if stimulus variables are present
        model = m.clone()
        p = model.get('stimulus').add_variable('period')
        p.set_unit(myokit.units.ms)
        p.set_rhs(1000)
        p.meta['oxmeta'] = 'membrane_stimulus_current_period'
        guess.add_embedded_protocol(model, protocol)

    def test_compatible_units(self):
        # Tests the (hidden) method to check if a unit is compatible with one
        # of a list of types
        A = myokit.units.A
        pA = myokit.units.pA
        AF = A / myokit.units.F
        Am2 = A / myokit.units.m**2

        units = [pA, AF, Am2]
        self.assertTrue(guess._compatible_units(pA, units))
        self.assertTrue(guess._compatible_units(A, units))
        self.assertTrue(guess._compatible_units(AF, units))
        self.assertTrue(guess._compatible_units(Am2, units))
        self.assertTrue(guess._compatible_units(123 * Am2, units))
        self.assertFalse(guess._compatible_units(myokit.units.m, units))
        self.assertFalse(guess._compatible_units(myokit.units.V, units))
        self.assertFalse(guess._compatible_units(None, units))

    def test_deep_deps(self):
        # Tests the (hidden) method to find all dependencies of a particular
        # variable.

        m = myokit.parse_model("""
            [[model]]
            c.v = -80

            [c]
            time = 0 bind time
            dot(v) = -i_ion
            i_ion = i1 + i2
            i1 = 5 * v
            i2 = x * v
                x = 13 / time
            z = 2 + y
            y = 3
            """)

        # Simple cases
        y, z = m.get('c.y'), m.get('c.z')
        d = guess._deep_deps(y)
        self.assertEqual(len(d), 0)
        d = guess._deep_deps(z)
        self.assertEqual(d[y], 1)
        self.assertEqual(len(d), 1)

        # Harder cases
        v, time, i_ion = m.get('c.v'), m.get('c.time'), m.get('c.i_ion')
        i1, i2, i2x = m.get('c.i1'), m.get('c.i2'), m.get('c.i2.x')
        d = guess._deep_deps(i2)
        self.assertEqual(d[i2x], 1)
        self.assertEqual(d[time], 2)
        self.assertEqual(len(d), 2)

        d = guess._deep_deps(v)
        self.assertEqual(d[i_ion], 1)
        self.assertEqual(d[i1], 2)
        self.assertEqual(d[i2], 2)
        self.assertEqual(d[i2x], 3)
        self.assertEqual(d[time], 4)
        self.assertEqual(len(d), 5)

    def test_distance_to_bound(self):
        # Tests the (hidden) method to calculate the distance to a bound
        # variable

        # Test case 1
        # Get distance to env.time, but only for variables that depend on
        # time but are otherwise constant
        m = myokit.parse_model("""
            [[model]]
            c.a = 1

            [env]
            time = 0 bind time

            [c]
            dot(a) = (b + 7) / a + c
            b = exp(a)
            c = 10 * b + d
            d = 1 + e
            e = 2 + f
            f = 3 / env.time
            g = 4 + e
            """)
        d = guess._distance_to_bound(m.get('env.time'))
        self.assertEqual(len(d), 4)
        self.assertEqual(d[m.get('c.d')], 3)
        self.assertEqual(d[m.get('c.e')], 2)
        self.assertEqual(d[m.get('c.f')], 1)
        self.assertEqual(d[m.get('c.g')], 3)
        self.assertEqual(m.get('env.time').binding(), 'time')

        # Test case 2
        m = myokit.parse_model("""
            [[model]]
            c.a = 1

            [env]
            time = 0 bind time

            [c]
            dot(a) = (b + 7) / a + c
            b = exp(a) + j
            c = 1 + env.time
            d = 2 / env.time
            e = 3 * env.time
            f = c + d
            g = j ^ d
            h = 1 - e
            i = cos(e)
            j = 2 * f
            """)
        d = guess._distance_to_bound(m.get('env.time'))
        self.assertEqual(len(d), 8)
        self.assertEqual(d[m.get('c.c')], 1)
        self.assertEqual(d[m.get('c.d')], 1)
        self.assertEqual(d[m.get('c.e')], 1)
        self.assertEqual(d[m.get('c.f')], 2)
        self.assertEqual(d[m.get('c.g')], 2)
        self.assertEqual(d[m.get('c.h')], 2)
        self.assertEqual(d[m.get('c.i')], 2)
        self.assertEqual(d[m.get('c.j')], 3)
        self.assertEqual(m.get('env.time').binding(), 'time')

        # Can only be called on bound variable
        self.assertRaisesRegex(
            ValueError, 'must be a bound variable',
            guess._distance_to_bound, m.get('c.a'))
        self.assertRaisesRegex(
            ValueError, 'must be a bound variable',
            guess._distance_to_bound, m.get('c.c'))

    def test_remove_nested(self):
        # Tests the method to remove nested variables

        m = myokit.parse_model("""
            [[model]]
            membrane.V = -80

            [membrane]
            t = 0 bind time
            dot(V) = 1 + x + y
                x = 1 + a + y + c
                    a = b
                        f = 9
                        b = 10 - f * g
                        g = 2
                    c = a + d
                    d = 4 * a
                y = 12
                in [mV]
                label membrane_potential
            """)
        m.validate()

        m2 = m.clone()
        v = m2.get('membrane.V')
        v.set_rhs(0)
        myokit.lib.guess._remove_nested(v)
        self.assertEqual(v.unit(), myokit.units.mV)
        self.assertEqual(len(v), 0)

    def test_membrane_capacitance(self):
        # Tests getting the membrane capacitance variable

        # Get by annotation
        m = myokit.parse_model("""
            [[model]]
            membrane.V = -80

            [membrane]
            t = 0 bind time
                oxmeta: it's me
            dot(V) = -1/Cm * i_ion
                oxmeta: membrane_capacitance
            i_ion = 4
            Cm = 1 [pF]
                in [pF]
            """)
        self.assertEqual(guess.membrane_capacitance(m).name(), 'V')

        # Get referenced in Vm
        m = myokit.parse_model("""
            [[model]]
            membrane.V = -80

            [membrane]
            t = 0 bind time
                oxmeta: it's me
            dot(V) = - i_ion
            i_ion = 4
            """)
        self.assertEqual(guess.membrane_capacitance(m).name(), 'i_ion')
        # Must be constant
        m.get('membrane.i_ion').set_rhs('4 * V')
        self.assertIsNone(guess.membrane_capacitance(m))

        # Get by name
        m = myokit.parse_model("""
            [[model]]
            membrane.V = -80

            [membrane]
            t = 0 bind time
            dot(V) = - i_ion
            i_ion = 4
            Cm = 3
                in [pF]
            """)
        c = m.get('membrane.Cm')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.rename('c')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.rename('c_m')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.rename('acap')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.rename('a_cap')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.rename('zzz')
        self.assertNotEqual(guess.membrane_capacitance(m), c)

        # Get by unit
        c.rename('Cm')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.set_unit(None)
        self.assertNotEqual(guess.membrane_capacitance(m), c)
        c.set_unit('pF/cm^2')
        self.assertEqual(guess.membrane_capacitance(m), c)
        c.set_unit('cm^2')
        self.assertEqual(guess.membrane_capacitance(m), c)

        # Bad unit cancels out point for name
        m = myokit.parse_model("""
            [[model]]

            [membrane]
            t = 0 bind time
            Cm = 3
                in [A]
            """)
        self.assertIsNone(guess.membrane_capacitance(m), c)

    def test_membrane_currents(self):
        # Tests getting a list of (outer) membrane currents

        # Test with annotated variable
        m = myokit.load_model('example')
        c = guess.membrane_currents(m)
        self.assertEqual(len(c), 6)
        c = [v.qname() for v in c]
        self.assertIn('ica.ICa', c)
        self.assertIn('ik.IK', c)
        self.assertIn('ik1.IK1', c)
        self.assertIn('ikp.IKp', c)
        self.assertIn('ina.INa', c)
        self.assertIn('ib.Ib', c)

        # Test with Vm
        i = m.label('cellular_current')
        i.set_label(None)
        c = guess.membrane_currents(m)
        self.assertEqual(len(c), 3)
        c = [v.qname() for v in c]
        self.assertIn('membrane.i_diff', c)
        self.assertIn('membrane.i_ion', c)
        self.assertIn('membrane.i_stim', c)

    def test_membrane_potential_1(self):
        # Test finding the membrane potential based on annotations

        m = myokit.parse_model("""
            [[model]]
            membrane.V = -80

            [membrane]
            time = 0 bind time
            dot(V) = 1 [mV/ms]
                in [mV]

            [c]
            x = 5 in [A]
            """)

        # Annotated with membrane_potential: return regardless of other factors
        x = m.get('c.x')
        x.set_label('membrane_potential')
        self.assertEqual(guess.membrane_potential(m), x)
        x.set_label(None)
        x.set_label('membrane_potentials')
        self.assertNotEqual(guess.membrane_potential(m), x)

        # Alternatively, use oxmeta annotation
        x.meta['oxmeta'] = 'membrane_voltage'
        self.assertEqual(guess.membrane_potential(m), x)
        del x.meta['oxmeta']
        self.assertNotEqual(guess.membrane_potential(m), x)

    def test_membrane_potential_2(self):
        # Test finding the membrane potential based on a scoring system

        # Must be in volts or None
        m = myokit.parse_model("""
            [[model]]
            membrane.V = -80

            [membrane]
            time = 0 bind time
            dot(V) = 1
                in [A]
            """)
        self.assertIsNone(guess.membrane_potential(m))
        v = m.get('membrane.V')
        v.set_unit('V')
        self.assertEqual(guess.membrane_potential(m), v)
        v.set_unit('MV')
        self.assertEqual(guess.membrane_potential(m), v)
        v.set_unit(None)
        self.assertEqual(guess.membrane_potential(m), v)

        # Prefer volts over none
        w = m.get('membrane').add_variable('Vm')
        w.set_rhs(1)
        w.promote(-80)
        w.set_unit('mV')
        self.assertEqual(guess.membrane_potential(m), w)
        w.set_unit(None)
        v.set_unit('kV')
        self.assertEqual(guess.membrane_potential(m), v)

        # Prefer common variable names
        m = myokit.parse_model("""
            [[model]]
            membrane.U = -80
            membrane.V = -80
            membrane.W = -80

            [membrane]
            time = 0 bind time
            dot(U) = 1
            dot(V) = 1
            dot(W) = 1
            """)
        self.assertEqual(guess.membrane_potential(m), m.get('membrane.V'))

        # Prefer common component names
        m = myokit.parse_model("""
            [[model]]
            a.V = -80
            membrane.V = -80
            c.V = -80

            [a]
            time = 0 bind time
            dot(V) = 1

            [membrane]
            dot(V) = 1

            [c]
            dot(V) = 1
            """)
        self.assertEqual(guess.membrane_potential(m), m.get('membrane.V'))
        # Synergy gives extra points
        m.get('a.V').set_unit('mV')
        self.assertEqual(guess.membrane_potential(m), m.get('membrane.V'))
        m.get('a.V').set_unit(None)
        # Component name works without variable name
        m.get('a.V').rename('W')
        m.get('membrane.V').rename('W')
        m.get('c.V').rename('W')
        self.assertEqual(guess.membrane_potential(m), m.get('membrane.W'))

        # Prefer states
        m = myokit.parse_model("""
            [[model]]
            b.V = -80

            [a]
            time = 0 bind time
            V = 1

            [b]
            dot(V) = 1

            [c]
            V = 1
            """)
        self.assertEqual(guess.membrane_potential(m), m.get('b.V'))

        # Use number of references as tie breaker
        m = myokit.parse_model("""
            [[model]]

            [a]
            time = 0 bind time
            V = 1
            x = V + b.V

            [b]
            V = 1

            [c]
            V = 1
            x = V + b.V
            """)
        self.assertEqual(guess.membrane_potential(m), m.get('b.V'))

        # Don't return really awful guesses
        m = myokit.parse_model("""
            [[model]]

            [a]
            time = 0 bind time
            x = 1
            y = 2
            z = 3
            """)
        self.assertIsNone(guess.membrane_potential(m))

    def test_remove_embedded_protocol(self):
        # Tests extracting an embedded protocol

        # Test without an offset
        model = myokit.parse_model("""
            [[model]]
            c.V = -80

            [c]
            time = 0 [ms]
                in [ms]
                bind time
            dot(V) = 0.1 * i_stim
            i_stim = if(time % period < duration, amplitude, 0)
                in [pA]
            duration = 0.3 [ms]
                in [ms]
            period = 500
            amplitude = -80 [pA]
            d = 3 [m]
            """)
        protocol = myokit.lib.guess.remove_embedded_protocol(model)
        self.assertIsInstance(protocol, myokit.Protocol)
        self.assertEqual(protocol.head().start(), 0)
        self.assertEqual(protocol.head().duration(), 0.3)
        self.assertEqual(protocol.head().period(), 500)

        # Test with an offset
        model = myokit.parse_model("""
            [[model]]
            c.V = -80

            [c]
            time = 0 [ms]
                in [ms]
                bind time
            dot(V) = 0.1 * i_stim
            i_stim = if((time - offset) % period < duration, amplitude, 0)
                in [pA]
            offset = 11 [ms]
                in [ms]
            duration = 0.3 [ms]
                in [ms]
            period = 500
            amplitude = -80 [pA]
            d = 3 [m]
            """)
        m2 = model.clone()
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsInstance(protocol, myokit.Protocol)
        self.assertEqual(protocol.head().start(), 11)
        self.assertEqual(protocol.head().duration(), 0.3)
        self.assertEqual(protocol.head().period(), 500)

        # Test various ways it can fail

        # Stimulus current not found
        m2 = model.clone()
        m2.get('c.V').set_rhs(1)
        m2.get('c').remove_variable(m2.get('c.i_stim'))
        m2.get('c').remove_variable(m2.get('c.amplitude'))
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsNone(protocol)

        # Duration not found
        m2 = model.clone()
        v = m2.get('c.i_stim')
        v.set_rhs('if((time - offset) % period < 0.4, 0, 0)')
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsNone(protocol)

        # Negative duration
        m2 = model.clone()
        m2.get('c.duration').set_rhs(-5)
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsNone(protocol)

        # Period not found
        m2 = model.clone()
        v = m2.get('c.i_stim')
        v.set_rhs('if((time - offset) < duration, 0, 0)')
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsNone(protocol)

        # Negative period
        m2 = model.clone()
        m2.get('c.period').set_rhs(-5)
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsNone(protocol)

        # Amplitude not found
        m2 = model.clone()
        v = m2.get('c.i_stim')
        v.set_rhs('if((time - offset) % period < duration, 0, 0)')
        protocol = myokit.lib.guess.remove_embedded_protocol(m2)
        self.assertIsNone(protocol)

        # Test recovering a protocol created by myokit
        m, p1, _ = myokit.load('example')
        myokit.lib.guess.add_embedded_protocol(m, p1)
        p2 = myokit.lib.guess.remove_embedded_protocol(m)
        self.assertIsInstance(p2, myokit.Protocol)
        self.assertEqual(p1.head().level(), p2.head().level())
        self.assertEqual(p1.head().start(), p2.head().start())
        self.assertEqual(p1.head().duration(), p2.head().duration())
        self.assertEqual(p1.head().period(), p2.head().period())
        self.assertEqual(p1.head().multiplier(), p2.head().multiplier())

    def test_stimulus_current_1(self):
        # Test finding the stimulus current based on annotations

        m = myokit.parse_model("""
            [[model]]
            c.V = -80

            [c]
            time = 0 bind time
            pace = 0 bind pace
            dot(V) = 0.1
            i_stim = pace * 10 [pA]
                in [pA]
            d = 3 [m]
            """)

        # Annotated with stimulus_current: return regardless of other factors
        x = m.get('c.d')
        x.set_label('stimulus_current')
        self.assertEqual(guess.stimulus_current(m), x)
        x.set_label(None)
        x.set_label('stimulus_currents')
        self.assertNotEqual(guess.stimulus_current(m), x)

        # Alternatively, use oxmeta annotation
        x.meta['oxmeta'] = 'membrane_stimulus_current'
        self.assertEqual(guess.stimulus_current(m), x)
        del x.meta['oxmeta']
        self.assertNotEqual(guess.stimulus_current(m), x)

    def test_stimulus_current_2(self):
        # Test finding the stimulus current using dependence on time or pace

        # Find variable furthest from time / pace (in A)
        m = myokit.parse_model("""
            [[model]]

            [t]
            pace = 0 bind pace
            time = 0 bind time
                in [s]
            a = time * 1 [A/s]
            b = 2 * a
            c = 3 * b
            d = 4 * c
            i_stim = 5 [A]
                in [A]
            """)
        a, b, c, d = [m.get(var) for var in ['t.a', 't.b', 't.c', 't.d']]
        self.assertEqual(guess.stimulus_current(m), d)

        # Prefer explicitly correct unit
        c.set_unit('A')
        self.assertEqual(guess.stimulus_current(m), c)
        c.set_unit(None)
        self.assertEqual(guess.stimulus_current(m), d)
        b.set_unit('A/F')
        self.assertEqual(guess.stimulus_current(m), b)
        d.set_unit('A/cm^2')
        self.assertEqual(guess.stimulus_current(m), d)

        # Disallow incorrect unit
        d.set_unit('m')
        self.assertEqual(guess.stimulus_current(m), b)
        b.set_unit(None)
        self.assertEqual(guess.stimulus_current(m), c)
        d.set_unit(None)
        self.assertEqual(guess.stimulus_current(m), d)

        # Prefer variables with known names
        m2 = m.clone()
        x = m2.get('t').add_variable('test')
        x.set_rhs('1 * a')
        self.assertEqual(guess.stimulus_current(m2), m2.get('t.d'))
        x = m2.get('t.i_stim')
        x.set_rhs('1 * a')
        self.assertEqual(guess.stimulus_current(m2), x)
        # Even if others have better units
        x.set_unit(None)
        m2.get('t.d').set_unit('A')
        self.assertEqual(guess.stimulus_current(m2), x)
        # Allow other names
        x.rename('istim')
        self.assertEqual(guess.stimulus_current(m2), x)
        x.rename('ist')
        self.assertEqual(guess.stimulus_current(m2), x)
        x.rename('i_st')
        self.assertEqual(guess.stimulus_current(m2), x)

        # This still works if using pace instead of time
        m.get('t.pace').set_binding(None)
        m.get('t.time').set_binding(None)
        m.get('t.time').set_binding('pace')
        self.assertEqual(guess.stimulus_current(m), d)

        # But not with other bindings
        m.get('t.time').set_binding(None)
        m.get('t.time').set_binding('plaice')
        self.assertNotEqual(guess.stimulus_current(m), d)

    def test_stimulus_current_3(self):
        # Test finding the stimulus current if it's a constant

        # Find variable furthest from time / pace (in A)
        m = myokit.parse_model("""
            [[model]]

            [t]
            time = 0 bind time
                in [s]
            i_stim = 5 [A]
                in [A]
            i_steam = 4 [A]
                in [A]
            """)

        # Correct name and unit
        i = m.get('t.i_stim')
        self.assertEqual(guess.stimulus_current(m), i)

        # Unit wrong or name wrong: don't return
        i.set_unit('m')
        self.assertIsNone(guess.stimulus_current(m))

        # Other allowed units
        i.set_unit('pA')
        self.assertEqual(guess.stimulus_current(m), i)
        i.set_unit('A/kF')
        self.assertEqual(guess.stimulus_current(m), i)
        i.set_unit('A/hm^2')
        self.assertEqual(guess.stimulus_current(m), i)
        i.set_unit(None)
        self.assertEqual(guess.stimulus_current(m), i)
        i.set_unit('mV')
        self.assertIsNone(guess.stimulus_current(m))

        # Other allowed names
        i.set_unit('pA')
        self.assertEqual(guess.stimulus_current(m), i)
        i.rename('istim')
        self.assertEqual(guess.stimulus_current(m), i)
        i.rename('ist')
        self.assertEqual(guess.stimulus_current(m), i)
        i.rename('i_st')
        self.assertEqual(guess.stimulus_current(m), i)
        i.rename('ii_st')
        self.assertIsNone(guess.stimulus_current(m))

    def test_stimulus_current_info_1(self):
        # Tests guessing of stimulus info based on annotations

        # Guess all using oxmeta labels
        m = myokit.parse_model("""
            [[model]]
            c.v = -80

            [c]
            t = 18 / 2
                oxmeta: membrane_stimulus_current
            a = 0 bind time
                oxmeta: membrane_stimulus_current_period
            dot(v) = -a
                oxmeta: membrane_stimulus_current_amplitude
            b = a / v
                oxmeta: membrane_stimulus_current_offset
            c = t
                oxmeta: membrane_stimulus_current_duration
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.t'))
        self.assertEqual(i['amplitude'], m.get('c.v'))
        self.assertEqual(i['duration'], m.get('c.c'))
        self.assertEqual(i['period'], m.get('c.a'))
        self.assertEqual(i['offset'], m.get('c.b'))
        self.assertIsNone(i['amplitude_expression'])

        # Try one without a label
        m.get('c.v').meta['oxmeta'] = 'monkey'
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['amplitude'])
        self.assertIsNotNone(i['amplitude_expression'])

        # Stimulus current can have a myokit label
        m.get('c.v').meta['oxmeta'] = 'membrane_stimulus_current_amplitude'
        m.get('c.t').set_label('stimulus_current')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.t'))
        self.assertEqual(i['amplitude'], m.get('c.v'))
        self.assertEqual(i['duration'], m.get('c.c'))
        self.assertEqual(i['period'], m.get('c.a'))
        self.assertEqual(i['offset'], m.get('c.b'))
        self.assertIsNone(i['amplitude_expression'])
        del m.get('c.t').meta['oxmeta']
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.t'))
        self.assertEqual(i['amplitude'], m.get('c.v'))
        self.assertEqual(i['duration'], m.get('c.c'))
        self.assertEqual(i['period'], m.get('c.a'))
        self.assertEqual(i['offset'], m.get('c.b'))
        self.assertIsNone(i['amplitude_expression'])

        # No current? no info
        m.get('c.t').set_label(None)
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['current'])
        self.assertIsNone(i['amplitude'])
        self.assertIsNone(i['duration'])
        self.assertIsNone(i['period'])
        self.assertIsNone(i['offset'])
        self.assertIsNone(i['amplitude_expression'])

    def test_stimulus_current_info_amplitude(self):
        # Tests guessing of stimulus amplitude

        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = 4 * if((t - s) % p < d, a, 0 [pA])
                in [pA]
            s = 10 [s] in [s]
            p = 1000 [s] in [s]
            d = 0.5 [s] in [s]
            a = 5 [pA] in [pA]
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.i_stim'))

        # Guess from correct unit or None
        a = m.get('c.a')
        self.assertEqual(i['amplitude'], a)
        a.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['amplitude'], a)

        # Reject incompatible unit
        a.set_unit('pF')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['amplitude'])
        a.set_unit(None)

        # Prefer explicit current over None
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = if((t - s) % p < d, a, 0 [pA])
                in [pA]
            s = 10 [s] in [s]
            p = 1000 [s] in [s]
            d = 0.5 [s] in [s]
            a = b
            b = 5 in [pA]
            """)
        b = m.get('c.b')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['amplitude'], b)
        b.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['amplitude'], b)

        # Guessing from name
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = 2 + if((t - s) % p < d, a, 0 [pA])
                in [pA]
            s = 10 [s] in [s]
            p = 1000 [s] in [s]
            d = 0.5 [s] in [s]
            a = b in [pA]
            b = c in [pA]
            c = 7 [pA] in [pA]
            """)
        b = m.get('c.b')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.i_stim'))
        self.assertNotEqual(i['amplitude'], b)
        b.rename('amplitude')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['amplitude'], b)
        b.rename('zamplitudez')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['amplitude'], b)
        b.rename('umplitudo')
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['amplitude'], b)
        b.rename('current_zamplitudez')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['amplitude'], b)
        b.rename('current_zzz')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['amplitude'], b)
        b.rename('umplitudo')
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['amplitude'], b)

        # Guessing from distance
        self.assertEqual(i['amplitude'], m.get('c.a'))

        # Can't already be used
        m.get('c.a').meta['oxmeta'] = 'membrane_stimulus_current_offset'
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['amplitude'], m.get('c.a'))

    def test_stimulus_current_info_amplitude_expression(self):
        # Tests guessing of stimulus amplitude expression

        # Expression is not set if amplitude can be determined
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = amplitude
                in [pA]
            amplitude = 1 [pA]
                in [pA]
            """)
        i = guess.stimulus_current_info(m)
        self.assertIsNotNone(i['amplitude'])
        self.assertIsNone(i['amplitude_expression'])

        # Simple constant (can be zero)
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = 0 [pA]
                in [pA]
            """)
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['amplitude'])
        self.assertEqual(str(i['amplitude_expression']), '0 [pA]')

        # More complicated constant
        i_stim = m.get('c.i_stim')
        r = '1 + 2 + 3 / sqrt(4)'
        i_stim.set_rhs(r)
        i = guess.stimulus_current_info(m)
        self.assertEqual(str(i['amplitude_expression']), r)

        # Non-zero part of an if (and ignore time variable)
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 bind time
            i_stim = if(t < 5, 1, 0)
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(str(i['amplitude_expression']), '1')

        # Non-zero part of an if
        i_stim = m.get('c.i_stim')
        i_stim.set_rhs('if(t < 5, 0, 2)')
        i = guess.stimulus_current_info(m)
        self.assertEqual(str(i['amplitude_expression']), '2')
        i_stim.set_rhs('1 - if(t < 5, 0, 0)')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['amplitude_expression'])

        # Non-zero part of a piecewise
        i_stim.set_rhs('piecewise(t < 5, 0, t < 10, 0, 0)')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['amplitude_expression'])
        i_stim.set_rhs('1 - piecewise(t < 5, 3, t < 10, 0, 0)')
        i = guess.stimulus_current_info(m)
        self.assertEqual(str(i['amplitude_expression']), '3')
        i_stim.set_rhs('2 + piecewise(t < 5, 0, t < 10, 2 / 4, 0)')
        i = guess.stimulus_current_info(m)
        self.assertEqual(str(i['amplitude_expression']), '2 / 4')
        i_stim.set_rhs('3 / piecewise(t < 5, 0, t < 10, 0, -18 / 2.4)')
        i = guess.stimulus_current_info(m)
        self.assertEqual(str(i['amplitude_expression']), '-18 / 2.4')

        # No suitable expression
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 bind time
            i_stim = t * 3
            """)
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['amplitude'])
        self.assertIsNone(i['amplitude_expression'])

    def test_stimulus_current_info_duration(self):
        # Tests guessing of stimulus duration

        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = if((t - s) % p < d, a, 0 [pA])
                in [pA]
            s = 10 [s] in [s]
                oxmeta: membrane_stimulus_current_offset
            p = 1000 [s] in [s]
                oxmeta: membrane_stimulus_current_period
            d = 0.5 [s] in [s]
            a = 5 [pA] in [pA]
                oxmeta: membrane_stimulus_current_amplitude
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.i_stim'))

        # Note: Having labels above allows the method to exit early. We can't
        # test this mini-optimisation, but it does show up in coverage

        # Guess from name + (correct unit or None)
        v = m.get('c.d')
        self.assertIsNone(i['duration'], v)
        v.rename('durandurations')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['duration'], v)
        v.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['duration'], v)
        v.set_unit('s^2')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['duration'], v)
        v.set_unit('1')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['duration'], v)

        # Can't already be used
        v.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['duration'], v)
        m.get('c.s').meta['oxmeta'] = 'hello'
        v.meta['oxmeta'] = 'membrane_stimulus_current_offset'
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['duration'], v)

    def test_stimulus_current_info_offset(self):
        # Tests guessing of stimulus offset

        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = if((t - s) % p < d, a, 0 [pA])
                in [pA]
            s = 10 [s] in [s]
            p = 1000 [s] in [s]
            d = 0.5 [s] in [s]
            a = 5 [pA] in [pA]
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.i_stim'))

        # Guess from name + (correct unit or None)
        v = m.get('c.d')
        self.assertIsNone(i['offset'], v)
        v.rename('foffsetti')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['offset'], v)
        v.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['offset'], v)
        v.set_unit('s^2')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['offset'], v)
        v.set_unit('1')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['offset'], v)

        # Can't already be used
        v.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['offset'], v)
        v.meta['oxmeta'] = 'membrane_stimulus_current_period'
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['offset'], v)

    def test_stimulus_current_info_period(self):
        # Tests guessing of stimulus period

        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = if((t - s) % p < d, a, 0 [pA])
                in [pA]
            s = 10 [s] in [s]
            p = 1000 [s] in [s]
            d = 0.5 [s] in [s]
            a = 5 [pA] in [pA]
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.i_stim'))

        # Guess from name + (correct unit or None)
        v = m.get('c.d')
        self.assertIsNone(i['period'], v)
        v.rename('stim_periodic')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['period'], v)
        v.set_unit(None)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['period'], v)
        v.set_unit('s^2')
        i = guess.stimulus_current_info(m)
        self.assertIsNone(i['period'], v)

        # Period can be dimensionless too
        v.set_unit('1')
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['period'], v)

        # Can't already be used
        v.meta['oxmeta'] = 'membrane_stimulus_current_offset'
        i = guess.stimulus_current_info(m)
        self.assertNotEqual(i['duration'], v)

    def test_stimulus_current_info_period_duration_offset(self):
        # Tests what happens when a single variable matches multiple words

        # In this model, amplitude should be x, and at most 2 variables out of
        # duration, period, and offset can be set.
        m = myokit.parse_model("""
            [[model]]
            [c]
            t = 0 [s] bind time
                in [s]
            i_stim = if((t - x) % perioddurationoffset < x, 1, 0)
            x = 5
            perioddurationoffset = 2
                in [s]
            """)
        i = guess.stimulus_current_info(m)
        self.assertEqual(i['current'], m.get('c.i_stim'))
        self.assertEqual(i['amplitude'], m.get('c.x'))
        self.assertIsNone(i['amplitude_expression'])
        empties = [x for x in i.values() if x is None]
        self.assertEqual(len(empties), 3)

        # Don't match period
        x = m.get('c.perioddurationoffset')
        x.rename('durationoffset')
        i = guess.stimulus_current_info(m)
        empties = [k for k, v in i.items() if v is None]
        self.assertEqual(len(empties), 3)
        self.assertIn('period', empties)

        # Don't match duration
        x.rename('periodnoffset')
        i = guess.stimulus_current_info(m)
        empties = [k for k, v in i.items() if v is None]
        self.assertEqual(len(empties), 3)
        self.assertIn('duration', empties)

        # Don't match offset
        x.rename('durationperiod')
        i = guess.stimulus_current_info(m)
        empties = [k for k, v in i.items() if v is None]
        self.assertEqual(len(empties), 3)
        self.assertIn('offset', empties)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the lib.hh module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit
import myokit.lib.hh as hh

from myokit.tests import DIR_DATA


MODEL = """
[[model]]
membrane.V = -80
ikr.a = 4.5e-4
ikr.r = 0.15
ina.r = 0.15
binding.act = 1e-4
binding.rec = 0.56
binding.b = 1e-4

[engine]
time = 0 bind time
pace = 0 bind pace

[membrane]
dot(V) = 0
    label membrane_potential

[ikr]
use membrane.V
IKr = g * a * r^2 * (V - EK)
dot(a) = (inf - a) / tau
    inf = alpha * tau
    tau = 1 / (alpha + beta)
    alpha = 2e-4 * exp(+7e-2 * V)
    beta = 3e-5 * exp(-5e-2 * V)
dot(r) = alpha * (1 - r) - beta * r
    alpha = 8e-2 * exp(+3e-3 * V)
    beta = 5e-3 * exp(-1e-2 * V)
g = 3
EK = -85

[ina]
use membrane.V as V
dot(r) = alpha * (1 - r) - beta * r
    alpha = 8e-2 * exp(+3e-3 * V)
    beta = 5e-3 * exp(-1e-2 * V)

[binding]
use membrane.V
conc = 10 [nM]
dot(act) = k1 * (1 - act) - k2 * act
    k1 = 2e-4 * exp(7e-2 * V)
    k2 = 3e-5 * exp(-5e-2 * V)
dot(rec) = k4 * (1 - rec) - k3 * rec
    k3 = 9e-2 * exp(9e-3 * V)
    k4 = 5e-3 * exp(-3e-2 * V)
dot(b) = (inf - b) / tau
    inf = (kon * conc) * tau
    tau = 1 / (kon * conc + koff)
kon = 1e-4 [1/nM/ms]
koff = 1e-5 [1/ms]
g = 3
EK = -85
I = g * (1 - b) * act * rec * (V - EK)
"""


class HHDetectionTest(unittest.TestCase):
    """
    Tests methods to detect HH states.
    """

    def test_alpha_beta_form(self):
        # Test methods for working with alpha-beta form
        m = myokit.parse_model(MODEL)
        self.assertIsInstance(m, myokit.Model)
        v = m.get('membrane.V')
        a = m.get('ikr.a')
        r = m.get('ikr.r')

        # Test without v (detected from label)
        self.assertTrue(hh.has_alpha_beta_form(r))
        alph, beta = hh.get_alpha_and_beta(r)
        self.assertEqual(alph, m.get('ikr.r.alpha'))
        self.assertEqual(beta, m.get('ikr.r.beta'))
        self.assertFalse(hh.has_alpha_beta_form(a))
        self.assertIsNone(hh.get_alpha_and_beta(a))

        # Test with v as a state, without a label
        v.set_label(None)
        self.assertRaisesRegex(
            ValueError, 'Membrane potential must be given',
            hh.has_alpha_beta_form, r)
        self.assertTrue(hh.has_alpha_beta_form(r, v))
        alph, beta = hh.get_alpha_and_beta(r, v)
        self.assertEqual(alph, m.get('ikr.r.alpha'))
        self.assertEqual(beta, m.get('ikr.r.beta'))
        self.assertFalse(hh.has_alpha_beta_form(a, v))
        self.assertIsNone(hh.get_alpha_and_beta(a, v))

        # Test with v as a constant
        v.demote()
        v.set_rhs(-80)
        self.assertTrue(hh.has_alpha_beta_form(r, v))
        alph, beta = hh.get_alpha_and_beta(r, v)
        self.assertEqual(alph, m.get('ikr.r.alpha'))
        self.assertEqual(beta, m.get('ikr.r.beta'))
        self.assertFalse(hh.has_alpha_beta_form(a, v))
        self.assertIsNone(hh.get_alpha_and_beta(a, v))

        # Almost correct forms / different ways to fail
        def bad(e):
            r.set_rhs(e)
            self.assertFalse(hh.has_alpha_beta_form(r, v))
            r.set_rhs('alpha * (1 - r) - beta * r')
            self.assertTrue(hh.has_alpha_beta_form(r, v))

        def good(e):
            r.set_rhs(e)
            self.assertTrue(hh.has_alpha_beta_form(r, v))

        # r is not a state
        self.assertTrue(hh.has_alpha_beta_form(r, v))
        r.demote()
        self.assertFalse(hh.has_alpha_beta_form(r, v))
        r.promote(0)
        self.assertTrue(hh.has_alpha_beta_form(r, v))

        # Not a minus
        bad('alpha * (1 - r) + beta * r')

        # Minus, but terms aren't multiplies
        bad('alpha / (1 - r) - beta * r')
        bad('alpha * (1 - r) - beta / r')

        # Terms in multiplications can be switched
        good('(1 - r) * alpha - beta * r')
        good('(1 - r) * alpha - r * beta')
        good('alpha * (1 - r) - beta * r')

        # But the correct terms are required
        bad('alpha * (2 - r) - beta * r')
        bad('alpha^2 * (1 - r) - beta * r')
        bad('alpha * (1 - r) - beta * r^2')
        bad('alpha * (1 - r) - beta^2 * r')

        # Alpha and beta can't be states
        m2 = m.clone()
        m2.get('membrane.V').set_label('membrane_potential')
        m2.get('ikr.r').move_variable(
            m2.get('ikr.r.alpha'), m2.get('ikr'), 'ralpha')
        self.assertTrue(hh.has_alpha_beta_form(m2.get('ikr.r')))
        m2.get('ikr.ralpha').promote(1)
        self.assertFalse(hh.has_alpha_beta_form(m2.get('ikr.r')))
        m2.get('ikr.ralpha').demote()
        m2.get('ikr.r').move_variable(
            m2.get('ikr.r.beta'), m2.get('ikr'), 'rbeta')
        self.assertTrue(hh.has_alpha_beta_form(m2.get('ikr.r')))
        m2.get('ikr.rbeta').promote(1)

        # Alpha and beta can't depend on other states than v
        c = m.add_component('ccc')
        x = c.add_variable('vvv')
        x.set_rhs(1)
        x.promote(0)
        ralph = m.get('ikr.r.alpha').rhs()
        self.assertTrue(hh.has_alpha_beta_form(r, v))
        m.get('ikr.r.alpha').set_rhs('3 * ccc.vvv + V')
        self.assertFalse(hh.has_alpha_beta_form(r, v))
        m.get('ikr.r.alpha').set_rhs(ralph)
        self.assertTrue(hh.has_alpha_beta_form(r, v))
        ralph = m.get('ikr.r.beta').rhs()
        m.get('ikr.r.beta').set_rhs('2 + ccc.vvv - V')
        self.assertFalse(hh.has_alpha_beta_form(r, v))
        m.get('ikr.r.beta').set_rhs(ralph)
        self.assertTrue(hh.has_alpha_beta_form(r, v))

    def test_inf_tau_form(self):
        # Test methods for working with inf-tau form
        m = myokit.parse_model(MODEL)
        self.assertIsInstance(m, myokit.Model)
        v = m.get('membrane.V')
        a = m.get('ikr.a')
        r = m.get('ikr.r')

        # Test with v detected from label
        self.assertTrue(hh.has_inf_tau_form(a))
        inf, tau = hh.get_inf_and_tau(a)
        self.assertEqual(inf, m.get('ikr.a.inf'))
        self.assertEqual(tau, m.get('ikr.a.tau'))
        self.assertFalse(hh.has_inf_tau_form(r))
        self.assertIsNone(hh.get_inf_and_tau(r))

        # Test with v argument, no label
        v.set_label(None)
        self.assertRaisesRegex(
            ValueError, 'Membrane potential must be given',
            hh.has_inf_tau_form, a)
        self.assertTrue(hh.has_inf_tau_form(a, v))
        inf, tau = hh.get_inf_and_tau(a, v)
        self.assertEqual(inf, m.get('ikr.a.inf'))
        self.assertEqual(tau, m.get('ikr.a.tau'))
        self.assertFalse(hh.has_inf_tau_form(r, v))
        self.assertIsNone(hh.get_inf_and_tau(r, v))

        # Test with v as a constant
        v.demote()
        v.set_rhs(-80)
        self.assertTrue(hh.has_inf_tau_form(a, v))
        inf, tau = hh.get_inf_and_tau(a, v)
        self.assertEqual(inf, m.get('ikr.a.inf'))
        self.assertEqual(tau, m.get('ikr.a.tau'))
        self.assertFalse(hh.has_inf_tau_form(r, v))
        self.assertIsNone(hh.get_inf_and_tau(r, v))
        del r

        # a is not a state
        self.assertTrue(hh.has_inf_tau_form(a, v))
        a.demote()
        self.assertFalse(hh.has_inf_tau_form(a, v))
        a.promote(0)
        self.assertTrue(hh.has_inf_tau_form(a, v))

        # Almost correct forms / different ways to fail
        def bad(e):
            a.set_rhs(e)
            self.assertFalse(hh.has_inf_tau_form(a, v))
            a.set_rhs('(inf - a) / tau')
            self.assertTrue(hh.has_inf_tau_form(a, v))

        def good(e):
            a.set_rhs(e)
            self.assertTrue(hh.has_inf_tau_form(a, v))

        bad('(inf - a) / (1 + tau)')
        bad('(inf + a) / tau')
        bad('(inf - 1) / tau')
        bad('(1 - inf) / tau')
        bad('(inf - r) / tau')

        # Inf and tau can't be states
        m.get('ikr.a').move_variable(m.get('ikr.a.inf'), m.get('ikr'), 'ainf')
        m.get('ikr.a').move_variable(m.get('ikr.a.tau'), m.get('ikr'), 'atau')
        self.assertTrue(hh.has_inf_tau_form(a, v))
        m.get('ikr.ainf').promote(1)
        self.assertFalse(hh.has_inf_tau_form(a, v))
        m.get('ikr.ainf').demote()
        self.assertTrue(hh.has_inf_tau_form(a, v))
        m.get('ikr.atau').promote(1)
        self.assertFalse(hh.has_inf_tau_form(a, v))
        m.get('ikr.atau').demote()
        self.assertTrue(hh.has_inf_tau_form(a, v))

        # Inf and tau can't depend on other states than v
        c = m.add_component('ccc')
        x = c.add_variable('vvv')
        x.set_rhs(1)
        x.promote(0)
        ainf = m.get('ikr.ainf').rhs()
        m.get('ikr.ainf').set_rhs('2 + ccc.vvv * V')
        self.assertFalse(hh.has_inf_tau_form(a, v))
        m.get('ikr.ainf').set_rhs(ainf)
        self.assertTrue(hh.has_inf_tau_form(a, v))
        atau = m.get('ikr.atau').rhs()
        m.get('ikr.atau').set_rhs('3 * ccc.vvv * V')
        self.assertFalse(hh.has_inf_tau_form(a, v))
        m.get('ikr.atau').set_rhs(atau)
        self.assertTrue(hh.has_inf_tau_form(a, v))

    def test_convert_hh_states_to_inf_tau_form(self):
        # Tests conversion to inf-tau form
        m1 = myokit.parse_model(MODEL)
        self.assertTrue(hh.has_inf_tau_form(m1.get('ikr.a')))
        self.assertFalse(hh.has_inf_tau_form(m1.get('ikr.r')))

        # Rewrite model
        m2 = hh.convert_hh_states_to_inf_tau_form(m1)
        self.assertNotEqual(m1.code(), m2.code())
        self.assertTrue(hh.has_inf_tau_form(m1.get('ikr.a')))
        self.assertFalse(hh.has_inf_tau_form(m1.get('ikr.r')))
        self.assertTrue(hh.has_inf_tau_form(m2.get('ikr.a')))
        self.assertTrue(hh.has_inf_tau_form(m2.get('ikr.r')))

        # Test only r was affected
        self.assertEqual(m1.get('ikr.a').code(), m2.get('ikr.a').code())
        self.assertNotEqual(m1.get('ikr.r').code(), m2.get('ikr.r').code())

        # Test form
        self.assertEqual(m2.get('ikr.r').rhs().code(), '(inf - ikr.r) / tau')
        a = m2.get('ikr.r.alpha').eval()
        b = m2.get('ikr.r.beta').eval()
        inf = m2.get('ikr.r.inf').eval()
        tau = m2.get('ikr.r.tau').eval()
        self.assertAlmostEqual(inf, a / (a + b))
        self.assertAlmostEqual(tau, 1 / (a + b))

        # Test state values are similar
        self.assertAlmostEqual(
            m1.get('ikr.r').eval(),
            m2.get('ikr.r').eval()
        )

        # Second rewrite isn't necessary
        m3 = hh.convert_hh_states_to_inf_tau_form(m2)
        self.assertEqual(m2.code(), m3.code())

        # First argument must be a myokit model
        self.assertRaisesRegex(
            ValueError, 'must be a myokit.Model',
            hh.convert_hh_states_to_inf_tau_form, [])

        # Membrane potential given explicitly (not as label)
        m2 = m1.clone()
        v = m2.get('membrane.V')
        v.set_label(None)
        m3 = hh.convert_hh_states_to_inf_tau_form(m2, v)
        self.assertNotEqual(m2.code(), m3.code())
        # Note: the next methods are called with v from m2, not v from m3! But
        # this should still work as variables are .get() from the model.
        self.assertTrue(hh.has_inf_tau_form(m3.get('ikr.a'), v))
        self.assertTrue(hh.has_inf_tau_form(m3.get('ikr.r'), v))

        # Unknown membrane potential
        self.assertRaisesRegex(
            ValueError, 'Membrane potential must be given',
            hh.convert_hh_states_to_inf_tau_form, m2)

    def test_rush_larsen_conversion(self):
        # Tests methods for writing RL state updates
        m1 = myokit.parse_model(MODEL)
        m2 = hh.convert_hh_states_to_inf_tau_form(m1)

        # Test RL update method
        dt = myokit.Name('dt')
        rl = hh.get_rl_expression(m2.get('ikr.a'), dt)
        self.assertEqual(
            rl.code(), 'inf + (ikr.a - inf) * exp(-(str:dt / tau))')
        rl = hh.get_rl_expression(m2.get('ikr.r'), dt)
        self.assertEqual(
            rl.code(), 'inf + (ikr.r - inf) * exp(-(str:dt / tau))')

        # Test RL update is close to Euler for small dt
        m2.get('membrane.V').set_rhs(20)
        ikr = m2.get('ikr')
        dt = ikr.add_variable('dt')
        dt.set_rhs(1e-6)
        # Test for a
        a = m2.get('ikr.a')
        rl = hh.get_rl_expression(a, myokit.Name(dt))
        a1 = rl.eval()
        a2 = a.initial_value(True) + dt.eval() * a.rhs().eval()
        self.assertAlmostEqual(a1, a2)
        # And for r
        r = m2.get('ikr.r')
        rl = hh.get_rl_expression(r, myokit.Name(dt))
        r1 = rl.eval()
        r2 = r.initial_value(True) + dt.eval() * r.rhs().eval()
        self.assertAlmostEqual(r1, r2)

        # Dt must be an expression
        self.assertRaisesRegex(
            ValueError, 'must be a myokit.Expression',
            hh.get_rl_expression, a, 'dt')

        # Returns None if not in inf-tau form
        self.assertIsNone(hh.get_rl_expression(
            m1.get('ikr.r'), myokit.Name(dt)))


class HHModelTest(unittest.TestCase):
    """
    Tests the HHModel class.
    """

    def test_manual_creation(self):
        # Test manual creation of a HHModel.

        # Load model
        fname = os.path.join(DIR_DATA, 'lr-1991-fitting.mmt')
        model = myokit.load_model(fname)

        # Select a number of states and parameters
        states = [
            'ina.m',
            model.get('ina.h'),
            'ina.j',
        ]
        parameters = [
            'ina.p1',
            model.get('ina.p10'),
            model.get('ina.p20'),
            'ina.p30',
        ]
        current = 'ina.INa'

        # Create a current model
        m = hh.HHModel(model, states, parameters, current)
        hh.HHModel(model, states, parameters, model.get(current))

        # Check membrane potential
        self.assertEqual(m.membrane_potential(), 'membrane.V')

        # Check parameters
        parameters[1] = parameters[1].qname()
        parameters[2] = parameters[2].qname()
        for p in m.parameters():
            self.assertIn(p, parameters)

        # First argument not a model
        self.assertRaisesRegex(
            ValueError, 'must be a myokit.Model', hh.HHModel, 1, states)

        # State doesn't exist
        self.assertRaisesRegex(
            hh.HHModelError, 'Unknown state',
            hh.HHModel, model, states + ['bert'], parameters, current)

        # Variable isn't a state
        self.assertRaisesRegex(
            hh.HHModelError, 'not a state',
            hh.HHModel, model, states + ['ina.INa'], parameters, current)

        # State added twice
        self.assertRaisesRegex(
            hh.HHModelError, 'twice',
            hh.HHModel, model, states + ['ina.h'], parameters, current)

        # No parameters is allowed
        hh.HHModel(model, states, None, current)

        # Non-literal parameter
        self.assertRaisesRegex(
            hh.HHModelError, 'Unsuitable',
            hh.HHModel, model, states, parameters + ['ina.INa'], current)

        # Parameter added twice
        self.assertRaisesRegex(
            hh.HHModelError, 'Parameter listed twice',
            hh.HHModel,
            model, states, parameters + ['ina.p1'], current)

        # Unknown parameter
        self.assertRaisesRegex(
            hh.HHModelError, 'Unknown parameter', hh.HHModel,
            model, states, parameters + ['ina.p1000'], current)

        # Current is a state
        self.assertRaisesRegex(
            hh.HHModelError, 'Current variable can not be a state',
            hh.HHModel, model, states, parameters, 'ina.m')

        # Current is not a function of the states
        m2 = model.clone()
        hh.HHModel(m2, states, parameters, current)
        m2.get(current).set_rhs(0)
        self.assertRaisesRegex(
            hh.HHModelError, 'Current must be a function of',
            hh.HHModel, m2, states, parameters, current)

        # Vm not given in model
        m2 = model.clone()
        hh.HHModel(m2, states, parameters, current)
        m2.get('membrane.V').set_label(None)
        self.assertRaisesRegex(
            hh.HHModelError, 'potential must be specified',
            hh.HHModel, m2, states, parameters, current)
        hh.HHModel(m2, states, parameters, current, vm='membrane.V')

        # Vm can't be a parameter
        self.assertRaisesRegex(
            hh.HHModelError, 'list of parameters', hh.HHModel,
            m2, states, parameters, current, vm=parameters[0])

        # Vm can't be a state
        self.assertRaisesRegex(
            hh.HHModelError, 'list of states',
            hh.HHModel, m2, states, parameters, current, vm=states[0])

        # Vm can't be the current
        self.assertRaisesRegex(
            hh.HHModelError, 'be the current',
            hh.HHModel, m2, states, parameters, current, vm=current)

        # States must have inf-tau form or alpha-beta form
        m2 = model.clone()
        m2.get('ina.m').set_rhs('alpha * (1 - m) - beta * m + 2')
        self.assertRaisesRegex(
            hh.HHModelError, 'must have "inf-tau form" or "alpha-beta form"',
            hh.HHModel, m2, states, parameters, current)

        # Current doesn't depend on states
        m2 = model.clone()
        m2.get(current).set_rhs('p1 + 12')
        self.assertRaisesRegex(
            hh.HHModelError, 'urrent must be a function of',
            hh.HHModel, m2, states, parameters, current)

        # Check states corresponds to input
        states[1] = states[1].qname()
        for s in m.states():
            self.assertIn(s, states)

        # Unlisted states are clamped
        v = model.get('ina').add_variable('xyz')
        v.promote(1.23)
        v.set_rhs(10)
        m = hh.HHModel(model, states, parameters, current)
        for s in m.states():
            self.assertIn(s, states)

        # All bindings other than time are removed
        # (e.g. this works without errors)
        i = model.get(current)
        i.set_binding('hello')
        m = hh.HHModel(model, states, parameters, current)

        # Returned reduced model is shorter than input model
        model2 = m.reduced_model()
        self.assertIsInstance(model2, myokit.Model)
        self.assertTrue(len(model2.code()) < len(model.code()))

    def test_automatic_creation(self):
        # Create a HH model from a component.

        # Load model
        fname = os.path.join(DIR_DATA, 'lr-1991-fitting.mmt')
        model = myokit.load_model(fname)

        # Create a current model
        hh.HHModel.from_component(model.get('ina'))

        # Test partially automatic creation
        states = [
            'ina.m',
            model.get('ina.h'),
            'ina.j',
        ]
        hh.HHModel.from_component(model.get('ina'), states=states)

        parameters = [
            'ina.p1',
            model.get('ina.p10'),
            model.get('ina.p20'),
            'ina.p30',
        ]
        hh.HHModel.from_component(
            model.get('ina'), parameters=parameters)

        current = 'ina.INa'
        hh.HHModel.from_component(model.get('ina'), current=current)
        hh.HHModel.from_component(model.get('ina'), current=model.get(current))

        # No current --> This is allowed
        m2 = model.clone()
        m2.get('ina').remove_variable(m2.get('ina.INa'))
        hh.HHModel.from_component(m2.get('ina'))

        # Two currents
        m2 = model.clone()
        v = m2.get('ina').add_variable('INa2')
        v.set_rhs(m2.get('ina.INa').rhs().clone())
        self.assertRaisesRegex(
            hh.HHModelError,
            'more than one variable that could be a current',
            hh.HHModel.from_component, m2.get('ina'))

        # Explict vm
        m2 = model.clone()
        m2.get('membrane.V').set_label(None)
        hh.HHModel.from_component(model.get('ina'), vm='membrane.V')
        self.assertRaisesRegex(
            hh.HHModelError,
            'labeled as "membrane_potential"',
            hh.HHModel.from_component, m2.get('ina'))

    def test_hh_model_steady_state(self):
        # Test the method HHModel.steady_state().

        # Create model
        fname = os.path.join(DIR_DATA, 'lr-1991-fitting.mmt')
        model = myokit.load_model(fname)
        m = hh.HHModel.from_component(model.get('ina'))

        # Get steady state
        ss = np.array(m.steady_state())

        # Check that this is a valid steady state
        self.assertTrue(np.all(ss >= 0))
        self.assertTrue(np.all(ss <= 1))

        # Test if derivatives are zero
        for k, x in enumerate(['ina.m', 'ina.h', 'ina.j']):
            x = model.get(x)
            x.set_initial_value(ss[k])
            self.assertAlmostEqual(x.eval(), 0)

        # Test arguments
        m.steady_state(-20, m.default_parameters())
        self.assertRaisesRegex(
            ValueError, 'parameter vector size',
            m.steady_state, -20, [1])

    def test_manual_creation_with_v_independence(self):
        # Test with the v-independent states

        # Load model
        model = myokit.parse_model(MODEL)

        # Select a number of states and parameters
        states = ['binding.act', 'binding.rec', 'binding.b']
        parameters = ['binding.kon', 'binding.koff']

        # Create a HH model
        m = hh.HHModel(model, states, parameters)

        # Also select a current
        current = 'binding.I'
        m = hh.HHModel(model, states, parameters, current)

        # Test steady-state calculation doesn't fail
        m.steady_state(-80)

    def test_automatic_creation_with_v_independence(self):
        # Test with the v-independent states

        model = myokit.parse_model(MODEL)
        m = hh.HHModel.from_component(model.get('binding'))

        self.assertEqual(len(m.states()), 3)

    def test_initial_value_conversion(self):
        # Tests that initial value expressions are converted to floats

        model = myokit.parse_model(MODEL)
        model.get('ikr.a').set_initial_value('1 / sqrt(7)')
        model.get('ikr.r').set_initial_value('log(binding.koff, 10)')
        m = hh.HHModel.from_component(model.get('ikr'))
        x0 = m.default_state()
        self.assertEqual(len(x0), 2)
        self.assertIsInstance(x0[0], float)
        self.assertIsInstance(x0[1], float)
        self.assertAlmostEqual(x0[0], 0.377964473)
        self.assertAlmostEqual(x0[1], -5)


class AnalyticalSimulationTest(unittest.TestCase):
    """
    Tests :class:`myokit.lib.hh.AnalyticalSimulation`.
    """

    def test_create_and_run(self):
        # Test creating and running a simulation

        # Create a simulation
        fname = os.path.join(DIR_DATA, 'lr-1991-fitting.mmt')
        model = myokit.load_model(fname)
        m = hh.HHModel.from_component(model.get('ina'))

        # Bad constructors
        self.assertRaisesRegex(
            ValueError, 'HHModel', hh.AnalyticalSimulation, 1)
        self.assertRaisesRegex(
            ValueError, 'Protocol', hh.AnalyticalSimulation, m, 1)

        # Create properly
        s = hh.AnalyticalSimulation(m)

        # Times to evaluate at
        times = np.linspace(0, 100, 5)

        # Voltages to test at
        voltages = np.arange(-70, 0, 30)

        # Generate traces with "solve" method
        state = s.state()
        dstate = s.default_state()
        for v in voltages:
            s.set_membrane_potential(v)
            x, i = s.solve(times)

        # Solve shouldn't change the state
        self.assertEqual(state, s.state())
        self.assertEqual(dstate, s.default_state())

        # Run for a bit
        self.assertIsInstance(s.run(10), myokit.DataLog)

        # Run can append to log
        d = s.run(10)
        n = len(d['engine.time'])
        e = s.run(1, log=d)
        self.assertIs(d, e)
        self.assertTrue(len(d['engine.time']) > n)

        # Run for zero duration
        d = s.run(0)
        self.assertEqual(len(d.time()), 0)
        d = s.run(0)
        self.assertEqual(len(d.time()), 0)

        # No current variable? Then current can't be calculated
        model2 = model.clone()
        model2.get('ina').remove_variable(model2.get('ina.INa'))
        m2 = hh.HHModel.from_component(model2.get('ina'))
        s2 = hh.AnalyticalSimulation(m2)
        # But simulation still works
        self.assertIsInstance(s2.run(10), myokit.DataLog)
        del model2, m2, s2

        # Create protocol

        # Protocol times: prep, step, post, full
        tprep = 2800
        tstep = 15
        tpost = 0

        # Step voltages
        vhold = -80
        vlo = -140
        vhi = 100
        res = 50
        v = np.arange(vlo, vhi + res, res)
        p = myokit.pacing.steptrain(v, vhold, tprep, tstep, tpost)
        t = p.characteristic_time()

        # Create simulation with protocol (set_protocol is not supported)
        s = hh.AnalyticalSimulation(m, p)

        # Membrane potential and protocol can't be used simultaneously
        self.assertRaisesRegex(
            Exception, 'cannot be set if', s.set_membrane_potential, -80)

        # Pre should change the state and default state
        state = s.state()
        dstate = s.default_state()
        s.pre(tprep + tstep)
        self.assertNotEqual(state, s.state())
        self.assertNotEqual(dstate, s.default_state())
        self.assertRaises(ValueError, s.pre, -1)

        # Run should change the state, not the default state
        state = s.state()
        dstate = s.default_state()
        d = s.run(t)
        self.assertNotEqual(state, s.state())
        self.assertEqual(dstate, s.default_state())
        self.assertRaisesRegex(ValueError, 'Duration', s.run, -1)
        self.assertRaisesRegex(
            ValueError, 'Log interval', s.run, 1, log_interval=-1)
        d['hello'] = [1, 2, 3]
        self.assertRaisesRegex(ValueError, 'extra keys', s.run, 1, log=d)
        del d['hello']
        del d[next(iter(d.keys()))]
        self.assertRaisesRegex(ValueError, 'missing', s.run, 1, log=d)

        # Reset should reset the state
        s.reset()
        self.assertEqual(state, s.state())

        # Run can append to log
        d = s.run(10)
        n = len(d['engine.time'])
        e = s.run(1, log=d)
        self.assertIs(d, e)
        self.assertTrue(len(d['engine.time']) > n)

        # Run for zero duration
        d = s.run(0)
        self.assertEqual(len(d.time()), 0)
        d = s.run(0)
        self.assertEqual(len(d.time()), 0)

    def test_analytical_simulation_properties(self):
        # Test basic get/set methods of analytical simulation.

        # Create a simulation
        fname = os.path.join(DIR_DATA, 'lr-1991-fitting.mmt')
        model = myokit.load_model(fname)
        m = hh.HHModel.from_component(model.get('ina'))
        s = hh.AnalyticalSimulation(m)

        # membrane potential
        self.assertEqual(
            s.membrane_potential(), model.get('membrane.V').eval())
        s.set_membrane_potential(10)
        self.assertEqual(s.membrane_potential(), 10)

        # Parameter values
        p = list(range(len(s.parameters())))
        self.assertNotEqual(p, s.parameters())
        s.set_parameters(p)
        self.assertEqual(p, s.parameters())
        self.assertRaises(ValueError, s.set_parameters, p[:-1])

        # Change parameter with set_constant
        p[3] += 1
        self.assertNotEqual(p, s.parameters())
        s.set_constant(m.parameters()[3], p[3])
        self.assertEqual(p, s.parameters())

        # State
        state = np.zeros(len(s.state()))
        state[0] = 0.3
        state[1] = 0.5
        self.assertNotEqual(list(state), list(s.state()))
        s.set_state(state)
        self.assertEqual(list(state), list(s.state()))
        self.assertRaisesRegex(
            ValueError, 'Wrong size', s.set_state, state[:-1])
        state[0] = -.1
        self.assertRaisesRegex(
            ValueError, 'must be in the range', s.set_state, state)
        state[0] = 0
        state[1] = 1.1
        self.assertRaisesRegex(
            ValueError, 'must be in the range', s.set_state, state)

        # Default state
        dstate = np.zeros(len(s.default_state()))
        dstate[0] = 0.3
        dstate[1] = 0.5
        self.assertNotEqual(list(dstate), list(s.default_state()))
        s.set_default_state(dstate)
        self.assertEqual(list(dstate), list(s.default_state()))
        self.assertRaisesRegex(
            ValueError, 'Wrong size', s.set_default_state, dstate[:-1])
        dstate[0] = -.1
        self.assertRaisesRegex(
            ValueError, 'must be in the range', s.set_default_state, dstate)
        state[0] = 0
        state[1] = 1.1
        self.assertRaisesRegex(
            ValueError, 'must be in the range', s.set_default_state, dstate)

    def test_against_cvode(self):
        # Validate against a cvode sim.

        # Get a model
        fname = os.path.join(DIR_DATA, 'lr-1991-fitting.mmt')
        model = myokit.load_model(fname)
        model.get('membrane.V').set_binding('pace')

        # Create a protocol
        vs = [-30, -20, -10]
        p = myokit.pacing.steptrain(
            vsteps=vs,
            vhold=-120,
            tpre=8,
            tstep=2,
            tpost=0)
        t = p.characteristic_time()

        # Run an analytical simulation
        dt = 0.01
        m = hh.HHModel.from_component(model.get('ina'))
        s1 = hh.AnalyticalSimulation(m, p)
        d1 = s1.run(t, log_interval=dt).npview()

        s2 = myokit.Simulation(model, p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(t, log_interval=dt).npview()

        # Test protocol output is the same
        e = np.abs(d1['membrane.V'] - d2['membrane.V'])
        self.assertEqual(np.max(e), 0)

        # Test current output is very similar
        e = np.abs(d1['ina.INa'] - d2['ina.INa'])
        self.assertLess(np.max(e), 2e-4)

    def test_against_cvode_v_independent(self):

        # Test with v-independent states
        model = myokit.parse_model(MODEL)
        model.get('membrane.V').set_rhs(-80)
        model.get('membrane.V').demote()
        model.binding('pace').set_binding(None)
        model.get('membrane.V').set_binding('pace')

        # Create a protocol
        vs = [-30, -20, -10]
        p = myokit.pacing.steptrain(
            vsteps=vs,
            vhold=-120,
            tpre=8,
            tstep=2,
            tpost=0)
        t = p.characteristic_time()

        # Run an analytical simulation
        dt = 0.01
        m = hh.HHModel.from_component(model.get('binding'))
        s1 = hh.AnalyticalSimulation(m, p)
        d1 = s1.run(t, log_interval=dt).npview()

        s2 = myokit.Simulation(model, p)
        s2.set_tolerance(1e-8, 1e-8)
        d2 = s2.run(t, log_interval=dt).npview()

        # Test protocol output is the same
        e = np.abs(d1['membrane.V'] - d2['membrane.V'])
        self.assertEqual(np.max(e), 0)

        # Test current output is very similar
        e = np.abs(d1['binding.I'] - d2['binding.I'])
        self.assertLess(np.max(e), 2e-4)


if __name__ == '__main__':
    unittest.main()



#
# Parser for .mmt files
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import re

from collections import OrderedDict

import myokit

from myokit import ParseError, ProtocolParseError


def parse(source):
    """
    Parses strings in ``mmt`` format and returns a tuple ``(model,
    protocol, embedded script)``.

    ``None`` may be returned for any segment not appearing in the source.

    The source to parse can be given as a plain string, a sequence of lines
    or a stream of lines.
    """
    # Get raw stream
    raw = source
    if isinstance(raw, str):
        raw = raw.splitlines()
    raw = iter(raw)

    # Create tokenizer
    stream = Tokenizer(raw)

    # Start parsing
    model = protocol = script = None

    # Get segments
    token = expect(stream.peek(), SEGMENT_HEADER)
    if token[1][2:-2] == 'model':
        model = parse_model_from_stream(stream)
        token = expect(stream.peek(), [SEGMENT_HEADER, EOF])
    if token[0] == SEGMENT_HEADER and token[1][2:-2] == 'protocol':
        protocol = parse_protocol_from_stream(stream)
        token = expect(stream.peek(), [SEGMENT_HEADER, EOF])
    if token[0] == SEGMENT_HEADER and token[1][2:-2] == 'script':
        script = parse_script_from_stream(stream, raw)
        token = expect(stream.peek(), [EOF])
    if token[0] == SEGMENT_HEADER:
        expecting = ''
        if script is None:
            expecting = '[[script]]'
            if protocol is None:
                expecting = '[[protocol]] or ' + expecting
                if model is None:
                    expecting = '[[model]] or ' + expecting
        raise ParseError(
            'Invalid segment header', token[2], token[3],
            'Expecting ' + expecting)
    expect(next(stream), EOF)

    # Return
    return (model, protocol, script)


def parse_model(source):
    """
    Parses a model in ``mmt`` format.
    """
    # Get raw stream
    raw = source
    if isinstance(raw, str):
        raw = raw.splitlines()
    raw = iter(raw)

    # Parse and return
    stream = Tokenizer(raw)
    token = expect(stream.peek(), SEGMENT_HEADER)
    if token[1] != '[[model]]':
        raise ParseError(
            'Invalid segment header', token[2], token[3],
            'Expecting [[model]]')
    model = parse_model_from_stream(stream)
    expect(next(stream), EOF)
    return model


def parse_protocol(source):
    """
    Parses a protocol in ``mmt`` format.
    """
    # Get raw stream
    raw = source
    if isinstance(raw, str):
        raw = raw.splitlines()
    raw = iter(raw)

    # Parse and return
    stream = Tokenizer(raw)
    token = expect(stream.peek(), SEGMENT_HEADER)
    if token[1] != '[[protocol]]':
        raise ParseError(
            'Invalid segment header', token[2], token[3],
            'Expecting [[protocol]]')
    protocol = parse_protocol_from_stream(stream)
    expect(next(stream), EOF)
    return protocol


def parse_script(source):
    """
    Parses a script in ``mmt`` format.
    """
    # Get raw stream
    raw = source
    if isinstance(raw, str):
        raw = raw.splitlines()
    raw = iter(raw)

    # Parse and return
    stream = Tokenizer(raw)
    token = expect(stream.peek(), SEGMENT_HEADER)
    if token[1] != '[[script]]':
        raise ParseError(
            'Invalid segment header', token[2], token[3],
            'Expecting [[script]]')
    script = parse_script_from_stream(stream, raw)
    expect(next(stream), EOF)
    return script


def parse_state(state):
    """
    Parses an initial state declaration given in one of two formats::

        <var_name_1> = <value>
        <var_name_1> = <value>
        ...

    Blank lines, whitespace and indentation are ignored. All variable names
    must be given fully qualified. Parsed data from this format is returned as
    a dictionary mapping variable names to values.

    Alternatively, the input can be given as a list of floating point numbers
    separated by spaces, commas, semi-colons, line breaks etc::

        <value>
        <value>
        ...

    Parsed data from this input format is returned as a list of floating point
    numbers.
    """
    # Get raw stream
    raw = state
    if isinstance(raw, str):
        raw = raw.splitlines()
    raw = iter(raw)

    # Create tokenizer
    stream = Tokenizer(raw, check_indenting=False)

    # Start parsing
    if stream.peek()[0] == NAME:
        # Parse name = value format
        pairs = {}
        while stream.peek()[0] != EOF:
            name = expect(next(stream), NAME)[1]
            t = stream.peek()
            if t[0] != DOT:
                raise ParseError(
                    'Unexpected token, expecting "."', t[2], t[3],
                    'All variable names must be fully qualified: comp.var')
            expect(next(stream), DOT)
            name += '.' + expect(next(stream), NAME)[1]
            expect(next(stream), EQUAL)
            pairs[name] = parse_expression_stream(stream).eval()
            t = expect(next(stream), EOL)
        return pairs
    else:
        # Parse list-of-floats format
        expr = [EOF, INTEGER, FLOAT, PLUS, MINUS, PAREN_OPEN, FUNC_NAME]
        state = []
        while True:
            while stream.peek()[0] not in expr:
                next(stream)
            if stream.peek()[0] == EOF:
                return state
            state.append(parse_expression_stream(stream).eval())


def split(source):
    """
    Attempts to split the ``source`` into model, protocol and script segments.
    Any content before the [[model]] tag is discarded.
    """
    # Get raw stream
    raw = source
    if isinstance(raw, str):
        raw = raw.splitlines()
    raw = iter(raw)
    segments = ['', '', '']
    prot_or_script = ('[[protocol]]', '[[script]]')

    # Gather lines in stream
    lines = [line for line in raw]

    # Create glue string for joining up lines again
    glue = ''
    if lines:
        glue = '' if lines[0][-1:] == '\n' else '\n'

    # Create stream of lines
    i = 0

    # Try parsing model
    try:
        stream = Tokenizer(iter(lines))
        #while stream.peek()[0] == EOL: # tokenizer ignores blank lines
        #    next(stream)
        token = stream.peek()
        if token[0] == SEGMENT_HEADER and token[1] == '[[model]]':
            parse_model_from_stream(stream, syntax_only=True)
    except ParseError:
        pass

    # Next up should be a protocol or script, anything else is junk that will
    #  be added to the model code
    if stream:
        i = stream.peek()[2] - 1
    for line in lines[i:]:
        if line.strip() in prot_or_script:
            break
        i += 1

    # Get model code
    segments[0] = glue.join(lines[:i])

    # Nothing found? Then everything is model code. Return
    if i == len(lines):
        return tuple(segments)

    # Chop off first part, continue hopeful parsing
    lines = lines[i:]

    # This will work because the next line is [[protocol]] or [[script]]
    stream = Tokenizer(iter(lines))

    # Try parsing protocol
    try:
        #while stream.peek()[0] == EOL: # tokenizer ignores blank lines
        #    next(stream)
        token = stream.peek()
        if token[0] == SEGMENT_HEADER and token[1] == '[[protocol]]':
            parse_protocol_from_stream(stream)
            token = stream.peek()
            line = token[2] - 1 if token[0] == SEGMENT_HEADER else token[2]
            segments[1] = glue.join(lines[:line])
            lines = lines[line:]
            stream = Tokenizer(iter(lines))

    except ParseError as e:

        # Reasonable guess: the protocol extends at least until e.line
        # Try finding [[script]]
        i = e.line
        for line in lines[e.line:]:
            if line.strip() == '[[script]]':
                break
            i += 1
        segments[1] = glue.join(lines[:i])
        if i == len(lines):
            return tuple(segments)
        lines = lines[i:]

    # Remainder is script
    segments[2] = glue.join(lines)
    return tuple(segments)


def expect(token, expected):
    """
    Checks the type of the given ``token``. If the token type equals
    ``expected`` or is one of the types in the sequence ``expected``, the
    token is returned. If not, an ParseError is raised.
    """
    code, text, line, char = token
    # Expected token type can be integer or collection
    if type(expected) == int:
        expected = [expected]
    if code in expected:
        return token
    unexpected_token(token, expected)


def unexpected_token(token, expected):
    """
    Raises a formatted unexpected token error (ParseError).
    """
    code, text, line, char = token

    # Parse received token
    got = token_map[code]
    hide = [EOL, EOF, INDENT, DEDENT, AND, OR, NOT]
    if code not in hide:
        got += ' "' + text + '"'

    # Parse expected token(s) or string
    if type(expected) == int:
        expected = [expected]
    if not isinstance(expected, str):
        if len(expected) > 2:
            expected = 'one of [' \
                + ', '.join([token_str[i] for i in expected]) + ']'
        elif len(expected) == 2:
            expected = token_str[expected[0]] + ' or ' + token_str[expected[1]]
        else:
            expected = token_str[expected[0]]
    raise ParseError(
        'Syntax error', line, char,
        'Unexpected token ' + got + ' expecting ' + expected)


def reg_token(info, token, obj):
    """
    Registers this token for the given object.

    Creates a variable _token in the object, pointing to the token
    Creates an entry in the model's token list, containing a tuple (token, obj)
    """
    # Register token with object
    obj._token = token
    # Register (token, obj) in model
    if info and info.model:
        if token[2] not in info.model._tokens:
            d = info.model._tokens[token[2]] = {}
        else:
            d = info.model._tokens[token[2]]
        d[token[3]] = (token, obj)


class ParseInfo:
    def __init__(self):
        self.model = None
        self.initial_values = OrderedDict()
        self.initial_value_tokens = {}
        self.alias_map = {}
        self.user_functions = {}


def parse_model_from_stream(stream, syntax_only=False):
    """
    Parses a model.

    By setting the optional argument ``syntax_only`` to ``True`` the function
    can be used as a syntax check: In this case only the early stages of
    parsing are handled but variables aren't resolved, the model isn't
    validated etc. In this use-case, True is returned if the parsed text had no
    syntax errors.
    """
    # Create parse info object
    info = ParseInfo()

    # Parse header definition
    token = expect(next(stream), SEGMENT_HEADER)
    if token[1][2:-2] != 'model':
        raise ParseError(
            'Invalid segment header', token[2], token[3],
            'Expecting [[model]]')
    expect(next(stream), EOL)

    # Create model
    model = info.model = myokit.Model()
    reg_token(info, token, model)

    # Parse header data
    token = stream.peek()
    while token[0] in (NAME, FUNC_NAME, META_NAME):

        if token[0] == FUNC_NAME:

            # User function
            parse_user_function(stream, info)

        elif token[0] == META_NAME:

            # Meta data
            t = next(stream)
            meta_key = t[1].strip()
            if meta_key in model.meta:
                raise ParseError(
                    'Duplicate meta-data key', t[2], t[3],
                    'The meta-data key "' + meta_key + '" was already'
                    ' specified for this model.')
            expect(next(stream), COLON)
            t_next = expect(next(stream), [TEXT, EOL])
            meta_val = ''
            if t_next[0] == TEXT:
                meta_val = t_next[1].strip()
                expect(next(stream), EOL)
            model.meta[meta_key] = meta_val

        else:

            # Initial value
            t0 = next(stream)
            t1 = expect(next(stream), DOT)
            t2 = expect(next(stream), NAME)
            name = t0[1] + t1[1] + t2[1]
            if name in info.initial_values:
                raise ParseError(
                    'Duplicate initial value', t0[2], t0[3],
                    'A value for <' + name + '> was already specified.')
            expect(next(stream), EQUAL)
            expr = parse_proto_expression(stream)
            expect(next(stream), EOL)
            info.initial_values[name] = expr
            info.initial_value_tokens[name] = t0

        token = stream.peek()

    # Save order of state variables
    state_order = list(info.initial_values.keys())

    # Parse components
    while stream.peek()[0] == BRACKET_OPEN:
        parse_component(stream, info)
    expect(stream.peek(), (EOF, BRACKET_OPEN, SEGMENT_HEADER))

    # Syntax checking mode
    if syntax_only:
        return True

    # Order encountered tokens
    m = model._tokens
    model._tokens = {}
    for line in sorted(m.keys()):
        model._tokens[line] = {}
        for char in sorted(m[line].keys()):
            model._tokens[line][char] = m[line][char]

    # Resolve alias map
    resolve_alias_map_names(info)

    # Resolve variable references to objects
    # A.K.A. Parse proto expression into Myokit.Expression
    for var in model.variables(deep=True):
        if var._proto_rhs is not None:
            # No need to check for IllegalReferenceErrors here, since these
            # won't have been resolved in the first place.
            var.set_rhs(convert_proto_expression(var._proto_rhs, var, info))
        del var._proto_rhs

    # Resolve variable references in initial values
    for i, var in enumerate(model.states()):
        e = info.initial_values[var.qname()]
        expr = convert_proto_expression(e, model, info)
        var.set_initial_value(expr)
        del info.initial_values[var.qname()]

    # All initial variables must have been used
    for qname, e in info.initial_values.items():
        t = info.initial_value_tokens[qname]
        raise ParseError(
            'Unused initial value', t[2], t[3],
            'An unused initial value was found for "' + qname + '".')

    # Re-order the model state
    model.reorder_state(state_order)

    # Check the semantics of the model
    try:
        model.validate()
    except myokit.IntegrityError as e:
        t = e.token()
        if t:
            raise ParseError('IntegrityError', t[2], t[3], str(e), cause=e)
        else:   # pragma: no cover
            raise ParseError('IntegrityError', 0, 0, str(e), cause=e)

    # Return
    return model


def parse_user_function(stream, info):
    """
    Parses a user function.
    """
    # Parse name
    token, name, line, char = expect(next(stream), FUNC_NAME)

    # Parse argument list
    args = []
    expect(next(stream), PAREN_OPEN)
    token = expect(next(stream), [PAREN_CLOSE, NAME])
    while token[0] == NAME:
        args.append(token[1])
        token = expect(next(stream), [COMMA, PAREN_CLOSE])
        if token[0] == COMMA:
            token = expect(next(stream), NAME)

    # Parse template
    expect(next(stream), EQUAL)
    expr = convert_proto_expression(parse_proto_expression(stream, info))
    expect(next(stream), EOL)

    # Create user function
    try:
        info.model.add_function(name, args, expr)
    except (myokit.DuplicateFunctionName,
            myokit.DuplicateFunctionArgument,
            myokit.InvalidNameError,
            myokit.InvalidFunction) as e:
        raise ParseError(
            'Invalid function declaration', line, char, str(e), cause=e)


def parse_component(stream, info=None):
    """
    Parses a component
    """
    # Info should have model
    assert info is not None and info.model is not None

    # Parse component declaration
    expect(next(stream), BRACKET_OPEN)
    token = expect(next(stream), NAME)
    expect(next(stream), BRACKET_CLOSE)
    expect(next(stream), EOL)
    code, name, line, char = token
    try:
        component = info.model.add_component(name)
    except myokit.DuplicateName as e1:
        raise ParseError(
            'Duplicate component name', line, char, str(e1), cause=e1)
    except myokit.InvalidNameError as e2:  # pragma: no cover
        # Already caught by tokenizer
        raise ParseError(
            'Illegal component name', line, char, str(e2), cause=e2)
    reg_token(info, token, component)

    # Add alias map
    info.alias_map[component] = {}

    # Parse fields
    token = stream.peek()
    while token[0] not in [BRACKET_OPEN, SEGMENT_HEADER, EOF]:
        if token[0] == USE:
            # Alias
            parse_alias(stream, info, component)
        elif token[0] == META_NAME:
            # Meta data
            t = next(stream)
            meta_key = t[1].strip()
            if meta_key in component.meta:
                raise ParseError(
                    'Duplicate meta-data key', t[2], t[3],
                    'The meta-data key "' + meta_key + '" was already'
                    ' specified for this component.')
            expect(next(stream), COLON)
            t_next = expect(next(stream), [TEXT, EOL])
            meta_val = ''
            if t_next[0] == TEXT:
                meta_val = t_next[1].strip()
                expect(next(stream), EOL)
            component.meta[meta_key] = meta_val
        else:
            # Variable
            parse_variable(stream, info, component)
        token = stream.peek()


def parse_alias(stream, info, component):
    """
    Parses an alias definition
    ::

        alias_map = {
            comp1 : {
                alias_as_string : (use_token, comp_token, var_token),

    """
    # Ensure valid parse info and alias map
    amap = info.alias_map[component]

    # Parse
    token = expect(next(stream), [USE])
    while True:
        # Get referenced variable
        token_c = expect(next(stream), NAME)   # component
        token_v = next(stream)                 # dot
        if token_v[0] != DOT:
            raise ParseError(
                'Invalid reference', token_v[2], token_v[3],
                'Aliassed variables must be specified using their fully'
                ' qualified name (component.variable)')
        token_v = expect(next(stream), NAME)   # variable

        # Get reference name
        if stream.peek()[0] == AS:
            expect(next(stream), AS)
            code, name, line, char = expect(next(stream), NAME)
        else:
            code, name, line, char = token_v

        # Create alias
        amap[name] = (token, token_c, token_v)

        # Expecting end of line or next alias
        t_next = expect(next(stream), [COMMA, EOL])
        if t_next[0] == EOL:
            break


def parse_variable(stream, info, parent):
    """
    Parses a variable.

    In normal operation (when parsing models) the variable's expression is
    stored in Variable._proto_rhs until the full model has been parsed. This
    allows the rhs to be set with a fully resolved expression.
    """
    # List of tokens to register with the final variable
    toreg = []

    # Parse variable declaration
    token = expect(next(stream), [NAME, FUNC_NAME])
    toreg.append(token)
    code, name, line, char = token

    # Allow dot() function on lhs
    if code == FUNC_NAME:
        if name != 'dot':
            raise ParseError(
                'Illegal lhs', line, char, 'Only variable names or the dot()'
                ' function may appear on the left-hand side of an equation.')
        if type(parent) != myokit.Component:  # pragma: no cover
            # Cover pragma: Already caught by tokenizer
            raise ParseError(
                'Illegal variable declaration', line, char,
                'State variable declarations may not be nested.')
        toreg.append(expect(next(stream), PAREN_OPEN))
        token = expect(next(stream), NAME)
        toreg.append(expect(next(stream), PAREN_CLOSE))
        toreg.append(token)
        code, name, line, char = token
        is_state = True
    else:
        is_state = False

    # Check if name is already in use for an alias
    component = parent
    if not isinstance(component, myokit.Component):
        component = component.parent(myokit.Component)

    # Create variable
    try:
        var = parent.add_variable(name)
    except myokit.DuplicateName as e1:
        raise ParseError(
            'Duplicate variable name', line, char, str(e1), cause=e1)
    except myokit.InvalidNameError as e2:  # pragma: no cover
        # Cover pragma: already caught by tokenizer
        raise ParseError(
            'Illegal variable name', line, char, str(e2), cause=e2)

    # Register tokens
    for token in toreg:
        reg_token(info, token, var)
    del toreg

    # Set initial value for states
    if is_state:
        if not var.qname() in info.initial_values:
            raise ParseError(
                'Missing initial value', line, char,
                'No initial value was found for "' + var.qname() + '"')
        var.promote()

    # Parse definition, quick unit, bind, label and description syntax
    # These token must occur in a fixed order!
    unit = None
    token = expect(stream.peek(), (EQUAL, IN, BIND, LABEL, COLON, EOL))
    if token[0] == EQUAL:
        # Parse variable definition
        next(stream)
        # Save proto-expression for right-hand side
        var._proto_rhs = parse_proto_expression(stream, info)
        # Get rest of line
        token = expect(stream.peek(), (IN, BIND, LABEL, COLON, EOL))
    else:  # pragma: no cover
        # No right hand side set! <-- Caught by parser
        var._proto_rhs = None
    if token[0] == IN:
        # Parse variable unit
        next(stream)
        unit = parse_bracketed_unit(stream)
        token = expect(stream.peek(), (BIND, LABEL, COLON, EOL))
    if token[0] == BIND:
        # Parse bind
        parse_binding(stream, info, var)
        token = expect(stream.peek(), (LABEL, COLON, EOL))
    if token[0] == LABEL:
        # Parse label
        parse_label(stream, info, var)
        token = expect(stream.peek(), (COLON, EOL))
    if token[0] == COLON:
        # Colon found, rest of line is description
        next(stream)
        desc = expect(next(stream), TEXT)[1].strip()
        var.meta['desc'] = desc
    expect(next(stream), EOL)

    # Parse indented fields (temp vars, meta or unit)
    if stream.peek()[0] == INDENT:
        next(stream)
        while stream.peek()[0] != DEDENT:
            token = expect(stream.peek(), [NAME, META_NAME, IN, BIND, LABEL])
            code, name, line, char = token
            if token[0] == NAME:
                # Nested variable
                parse_variable(stream, info, var)
            elif token[0] == META_NAME:
                # Meta property
                meta_key = name.strip()
                if meta_key in var.meta:
                    raise ParseError(
                        'Duplicate meta-data key', line, char,
                        'The meta-data key "' + meta_key + '" was already'
                        ' specified for this variable.')
                next(stream)
                expect(next(stream), COLON)
                tnext = expect(next(stream), [TEXT, EOL])
                meta_val = ''
                if tnext[0] == TEXT:
                    meta_val = tnext[1].strip()
                    expect(next(stream), EOL)
                var.meta[meta_key] = meta_val
            elif token[0] == IN:
                if unit:
                    raise ParseError(
                        'Duplicate variable unit', line, char,
                        'Unit already specified for this variable.')
                # Variable unit
                next(stream)
                unit = parse_bracketed_unit(stream)
                expect(next(stream), EOL)
            elif token[0] == BIND:
                # Binding to external value
                parse_binding(stream, info, var)
                expect(next(stream), EOL)
            elif token[0] == LABEL:
                # Labelled as unique value
                parse_label(stream, info, var)
                expect(next(stream), EOL)
            else:   # pragma: no cover
                raise Exception('Unhandled case.')
            # Next line
            token = stream.peek()[0]
        expect(next(stream), DEDENT)

    # Set unit
    var.set_unit(unit)

    # Normal operation is to leave the ._proto_rhs untouched until the full
    # model has been parsed and variables have been resolved. For debugging, it
    # may be nice to resolve at this point already.
    #var.set_rhs(convert_proto_expression(var._proto_rhs))
    #del(var._proto_rhs)


def parse_binding(stream, info, var):
    """
    Parses the "bind" part of a variable definition.
    """
    token, name, line, char = expect(next(stream), BIND)

    # Get binding label
    label = expect(next(stream), NAME)[1]

    # Bind variable
    try:
        var.set_binding(label)
    except myokit.InvalidBindingError as e:  # pragma: no cover
        # Cover pragma: This is already caught when parsing the variable
        raise ParseError('Illegal binding', line, char, str(e), cause=e)


def parse_label(stream, info, var):
    """
    Parses the "label" part of a variable definition.
    """
    token, name, line, char = expect(next(stream), LABEL)
    # Get label
    label = expect(next(stream), NAME)[1]
    # Register label
    try:
        var.set_label(label)
    except myokit.InvalidLabelError as e:  # pragma: no cover
        # Cover pragma: This is already caught when parsing the variable
        raise ParseError('Illegal label', line, char, str(e), cause=e)


def resolve_alias_map_names(info):
    """
    Resolves the variables pointed to by the alias dictionary

    Updates the alias map in each component from a string:string dictionary to
    a string:variable dictionary and checks if
     - The variables are in another component
     - No two aliasses are tied to the same variable (within a single
      component)
    """
    for comp, amap in info.alias_map.items():
        for name, (t_use, t_comp, t_var) in amap.items():
            var_name = t_comp[1] + '.' + t_var[1]
            (var, sug, msg) = info.model.suggest_variable(var_name)
            if var is None:
                raise ParseError(
                    'Variable not found',
                    t_comp[2], t_comp[3], msg)
            try:
                comp.add_alias(name, var)
            except myokit.DuplicateName as e:
                raise ParseError(
                    'Duplicate name error',
                    t_comp[2], t_comp[3], str(e), cause=e)
            except myokit.InvalidNameError as e2:  # pragma: no cover
                # Cover pragma: Already caught by the tokenizer
                raise ParseError(
                    'Illegal alias name',
                    t_comp[2], t_comp[3], str(e2), cause=e2)


def parse_unit(stream):
    """
    Parses a unit expression and returns a Unit object.
    """
    # Parse first unit in unit expression
    token = expect(next(stream), [NAME, INTEGER])
    if token[0] == NAME:
        try:
            unit = myokit.Unit.parse_simple(token[1])
        except KeyError as ke:
            raise ParseError(
                'Unit not recognized', token[2], token[3], str(ke), cause=ke)
        if stream.peek()[0] == POWER:
            next(stream)
            expo = expect(next(stream), (INTEGER, FLOAT, MINUS))[1]
            if expo == '-':
                unit **= -float(expect(next(stream), (INTEGER, FLOAT))[1])
            else:
                unit **= float(expo)
    else:
        x = float(token[1])
        if x != 1.0:
            raise ParseError(
                'Invalid unit specification', token[2], token[3],
                'Unit specification must start with unit name or "1"')
        unit = myokit.Unit.parse_simple('1')

    # Parse remaining units (* or /)
    while stream.peek()[0] in [STAR, SLASH]:
        op = next(stream)[0]
        token = expect(next(stream), NAME)
        try:
            part = myokit.Unit.parse_simple(token[1])
        except KeyError as ke:
            raise ParseError(
                'Unit not recognized', token[2], token[3], str(ke), cause=ke)
        if stream.peek()[0] == POWER:
            next(stream)
            expo = expect(next(stream), (INTEGER, FLOAT, MINUS))[1]
            if expo == '-':
                part **= -float(expect(next(stream), (INTEGER, FLOAT))[1])
            else:
                part **= float(expo)
        if op == STAR:
            unit *= part
        else:
            unit /= part

    # Parse multiplier
    if stream.peek()[0] == PAREN_OPEN:
        next(stream)
        token = stream.peek()
        e = parse_expression_stream(stream)
        if not e.is_literal():
            raise ParseError(
                'Invalid unit multiplier', token[2], token[3],
                'Unit multipliers cannot contain variables.')
        unit *= e.eval()
        expect(next(stream), PAREN_CLOSE)

    return unit


def parse_unit_string(string):
    """
    Parses string data into a :class:`myokit.Unit`.
    """
    s = Tokenizer(string)
    e = parse_unit(s)
    expect(next(s), EOL)
    expect(next(s), EOF)
    # Note: EOF is a python thing, not an actual character, so checking that
    # an EOF is next is equivalent to checking that there are no further
    # characters.
    return e


def parse_bracketed_unit(stream):
    """
    Parses a unit wrapped in brackets, catches the unit text and adds a fake
    UNIT token to the unit, allowing later maniplation.
    """
    unit_text = stream.start_catching()
    token = expect(next(stream), BRACKET_OPEN)
    unit = parse_unit(stream)
    expect(next(stream), BRACKET_CLOSE)
    unit_text = stream.stop_catching(unit_text)
    unit._token = (UNIT, unit_text, token[2], token[3])
    return unit


def parse_protocol_from_stream(stream):
    """
    Parses a protocol segment, requires a Tokenstream
    """
    # Number parsing function
    def parse_number(stream):
        negative = False
        t = expect(next(stream), (PLUS, MINUS, INTEGER, FLOAT))
        if t[0] == PLUS:
            t = expect(next(stream), (INTEGER, FLOAT))
        elif t[0] == MINUS:
            negative = True
            t = expect(next(stream), (INTEGER, FLOAT))
        number = int(t[1]) if t[0] == INTEGER else float(t[1])
        return -number if negative else number

    # Check segment header
    if stream.peek()[0] == SEGMENT_HEADER:
        token = next(stream)
        if token[1][2:-2] != 'protocol':
            raise ParseError(
                'Invalid segment header', token[2], token[3],
                'Expecting [[protocol]]')
        expect(next(stream), EOL)

    # Create protocol
    protocol = myokit.Protocol()

    # Parse lines
    t_last = None
    t_next = 0
    n = stream.peek()
    while n[0] not in (EOF, SEGMENT_HEADER):

        # Parse level
        v = parse_number(stream)

        # Parse starting time
        # Allow 'next' to mean "after the previous event ends"
        if stream.peek()[1] == 'next':
            next(stream)
            if t_next is None:
                raise ProtocolParseError(
                    'Invalid next', n[2], n[3],
                    'Unable to determine end of previous event, "next" cannot'
                    ' be used here.')
            else:
                t = t_next
        else:
            t = parse_number(stream)
        if t_last is None:
            t_last = t
        else:
            if t > t_last:
                t_last = t
            elif t == t_last:
                raise ProtocolParseError(
                    'Simultaneous stimuli', n[2], n[3],
                    'Stimuli may not occur at the same time.')
            else:
                raise ProtocolParseError(
                    'Non-consecutive stimuli', n[2], n[3],
                    'Stimuli must be listed in chronological order.')

        # Parse duration
        d = parse_number(stream)
        if d <= 0:
            raise ProtocolParseError(
                'Invalid duration', n[2], n[3],
                'The duration of a stimulus must be greater than zero.')

        # Parse period
        p = parse_number(stream)
        if p < 0:
            raise ProtocolParseError(
                'Negative period', n[2], n[3],
                'Stimulus cannot occur with a negative period.')

        # Parse multiplier
        r = int(parse_number(stream))
        if r < 0:
            raise ProtocolParseError(
                'Negative multiplier', n[2], n[3],
                'Stimulus cannot occur a negative number of times.')
        elif r > 0 and p == 0:
            raise ProtocolParseError(
                'Invalid multiplier', n[2], n[3],
                'Non-periodic event cannot occur more than once.')

        # Determine next event end
        if p == 0:
            t_next = t + d
        else:
            t_next = None

        # Parse end of line
        expect(next(stream), EOL)

        # Schedule event
        try:
            protocol.schedule(v, t, d, p, r)
        except myokit.ProtocolEventError as e:  # pragma: no cover
            # Cover pragma: This should already have been caught!
            raise ProtocolParseError(
                'Invalid event specification', n[2], 0, str(e))
        n = stream.peek()

    return protocol


def parse_script_from_stream(stream, raw_stream):
    """
    Parses a script segment
    """
    token = expect(next(stream), SEGMENT_HEADER)
    if token[1][2:-2] != 'script':
        raise ParseError(
            'Invalid segment header', token[2], token[3],
            'Expecting [[script]]')
    token = expect(stream.peek(), EOL)
    raw = []
    for line in raw_stream:
        raw.append(line)
    expect(next(stream), EOL)
    return ''.join(raw)


def strip_expression_units(model_text, skip_literals=True):
    """
    Takes the text of a valid model as input and returns a version stripped of
    any expression units. Variable units defined with ``in`` are preserved.
    Only the model part should be passed in, no script or protocol segments.

    By default, constants defined for variables whose RHS is a single number
    will keep their units. To disable this, set ``skip_literals=False``.

    This method will raise a :class:`myokit.ParseError` if the given code
    cannot be parsed to a valid model.
    """
    if isinstance(model_text, str):
        lines = model_text.splitlines()
    else:
        lines = model_text
    stream = Tokenizer(lines)
    m = parse_model_from_stream(stream)
    # Collect positions of fake unit tokens
    cuts = []
    toks = []
    for var in m.variables(deep=True):
        rhs = var.rhs()
        if skip_literals and type(rhs) == myokit.Number:
            continue
        for e in rhs.walk(allowed_types=myokit.Number):
            u = e.unit()
            if u is not None:
                assert u._token is not None
                token, text, line, char = u._token
                toks.append(u._token)
                # Lines start at 1, chars start at 0...
                cuts.append((line - 1, char, char + len(text)))
    # Create stripped version of lines
    if cuts:
        stripped = []
        cuts.sort()
        cuts = iter(cuts)
        cut = next(cuts)
        for k, line in enumerate(lines):
            if cut and cut[0] == k:
                # Gather non-unit parts of line
                line2 = ''
                offset = 0
                while cut and cut[0] == k:
                    x, char1, char2 = cut
                    # Adjust spaces
                    if line[char1 - 1:char1] == ' ':
                        # Remove leading space if:
                        #  - unit is followed by a space: 1 [ms] * 10
                        #  - unit is followed by a bracket: (1 + 1 [ms])
                        #  - unit is at end of line: 1 [ms]
                        if line[char2:char2 + 1] in (' ', ')', ''):
                            char1 -= 1
                    line2 += line[offset:char1]
                    offset = char2
                    try:
                        cut = next(cuts)
                    except StopIteration:
                        cut = None
                line = line2 + line[offset:]
            stripped.append(line)
        lines = stripped
    return '\n'.join(lines)


# Tokens
token_map = {}
token_str = {}


def add_token(name, nice):
    """
    Defines a token in the global namespace of this module

    Arguments:
    name
        The variable name
    nice
        A human-readable version of the variable name
    """
    code = 10 + len(token_map)
    globals()[name] = code
    token_map[code] = name
    token_str[code] = nice


add_token('EOL', 'End of line')
add_token('EOF', 'End of file')
add_token('INDENT', 'Indentation increase')
add_token('DEDENT', 'Indentation decrease')
add_token('WHITESPACE', 'Whitespace between tokens')
add_token('NAME', 'Name (function, variable etc)')
add_token('INTEGER', 'Integer')
add_token('FLOAT', 'Float')
add_token('TEXT', 'Text')
add_token('FUNC_NAME', 'Function name')
add_token('META_NAME', 'Meta property name')
add_token('PAREN_OPEN', 'Left parentheses "("')
add_token('PAREN_CLOSE', 'Right parentheses ")"')
add_token('BRACKET_OPEN', 'Left bracket "["')
add_token('BRACKET_CLOSE', 'Right bracket "]"')
add_token('SEGMENT_HEADER', 'Segment header "[[segment_name]]"')
add_token('EQUAL', 'Assignment "="')
add_token('COLON', 'Colon ":"')
add_token('DOT', 'Dot "."')
add_token('COMMA', 'Comma ","')
add_token('PLUS', 'Plus "+"')
add_token('MINUS', 'Minus "-"')
add_token('STAR', 'Star "*"')
add_token('SLASH', 'Slash "/"')
add_token('POWER', 'Power "^"')
add_token('REMAINDER', 'Remainder "%"')
add_token('QUOTIENT', 'Quotient "//"')
add_token('EQEQUAL', 'Equality "=="')
add_token('NOTEQUAL', 'Inequality "!="')
add_token('MORE', 'More ">"')
add_token('LESS', 'Less "<"')
add_token('MOREEQUAL', 'More than or equal ">="')
add_token('LESSEQUAL', 'Less than or equal "<="')
add_token('AND', 'and')
add_token('OR', 'or')
add_token('NOT', 'not')
add_token('IN', 'in')           # Unit
add_token('USE', 'use')         # Alias
add_token('AS', 'as')           # Alias
add_token('BIND', 'bind')       # External value
add_token('LABEL', 'label')     # Special value
add_token('UNIT', 'unit')       # Used to given Units a useful token.

# Reserved keywords
KEYWORD_MAP = {
    'and': AND,
    'or': OR,
    'not': NOT,
    'in': IN,
    'use': USE,
    'as': AS,
    'bind': BIND,
    'label': LABEL,
}
KEYWORDS = KEYWORD_MAP.keys()

# Tabsize is used to interpret the column a tab will move you to, this affects
#  the indenting rules
tabsize = 8

# Define token recognising regex
# The syntax r'' is used for raw strings
_rTOKEN = re.compile('|'.join([
    # Whitespace
    r'[ \t]+',
    # Function opening (must come before name in list)
    r'[a-zA-Z]\w*\(',
    # Meta property name (must come before name in list)
    r'([a-zA-Z]\w*:)+',
    # Names
    r'[a-zA-Z]\w*',
    # Floating point numbers
    r'(([0-9]*\.[0-9]+)|([0-9]+\.?[0-9]*))([eE][+-]?[0-9]+)?',
    # Integers
    r'[0-9]+',
    # Comparison 1
    r'[<>!=]=',
    # Comparison 2
    r'[<>]',
    # Quotient and remainder
    r'[/]{2}|[%]',
    # Operators, * must come first (regex reasons...)
    r'[*+-/^]',
    # Segments
    r'\[\[[a-zA-Z]{1}[a-zA-Z0-9_]*\]\]',
    # Parentheses, brackets
    r'[()[\]{}]',
    # Assignment
    r"[=:]",
    # Dot, comma
    r'[.,]',
    # Line joining backslash
    r'[\\]',
]))

# White space
# Note: In unicode, a line feed (\f) is treated as a newline. This means \f
# characters will be stripped out by splitlines(), just like \n, and don't need
# further handling: https://en.wikipedia.org/wiki/Newline#Unicode
_sWHITE = ' \t'
_rWHITE = re.compile(r'[ \t]*')

# Recognizable characters
_sEOL = '\n'
_sNUMBERS = '0123456789'

# Tell floats from integers
_rFloat = re.compile(r'[.eE]+')

# Map of single characters
_SINGLE = "[]<>=.,+-*/^%"
_SINGLE_MAP = [
    BRACKET_OPEN, BRACKET_CLOSE,
    LESS, MORE, EQUAL, DOT, COMMA,
    PLUS, MINUS, STAR, SLASH, POWER, REMAINDER]

# Map of equals-comparators (==, !=, >=, <=)
_COMPEQ = '=!><'
_COMPEQ_MAP = [EQEQUAL, NOTEQUAL, MOREEQUAL, LESSEQUAL]


class Tokenizer:
    """
    Takes a stream of lines as input and provides a stream interface returning
    tokens.

    Strips comments and concatenates lines with open parentheses (anything
    between round brackets is counted as a single line). Blank lines (or lines
    with nothing but comments) are ignored. Leading whitespace is converted
    to INDENT and DEDENT tokens, other whitespace is ignored.

    The tokens are returned as a 4-tuple with the following members:

    * The token type
    * The string parsed into a token
    * The line number the token was found on
    * The position in the line the token was found at

    The current token can be obtained using :meth:`current()`, the tokenizer
    can be advanced to the next token with :meth:`next()` and a preview of the
    next token can be obtained using :meth:`peek()`. In other words; ``peek()``
    and ``next()`` return the same token but ``next()`` also advances the
    stream while ``peek()`` does not.

    For memory efficiency, the given stream is read one line at a time and no
    reference to the read lines or returned tokens is kept. To obtain a list of
    all tokens returned between time A and B, the "catcher" mechanism can be
    used. For details, see :meth:`start_catching()`.

    For files that are to be parsed without indenting rules, the optional
    argument ``check_indenting`` can be set to False. In this mode, the
    tokenizer won't yield ``INDENT`` or ``DEDENT`` tokens.
    """
    def __init__(self, stream_of_lines, check_indenting=True):
        # Set next value and peek value
        self._next = None
        self._peek = None

        # At end of stream?
        self._has_last_value = False

        # Catchers and catcher handle index
        self._catchers = {}
        self._catcheri = 0

        # String given instead of stream of lines? Convert
        if isinstance(stream_of_lines, str):
            stream_of_lines = iter(stream_of_lines.splitlines())

        # Create tokenizer
        self._tokenizer = self._tizer(stream_of_lines, check_indenting)

        # Grab first token
        self._advance()

    #def __iter__(self):
    #    return self

    def _advance(self):
        """
        Advances to the next token.
        """
        self._next = self._peek
        for c in self._catchers.values():
            c.append(self._next[1])
        try:
            self._peek = next(self._tokenizer)
            while self._peek[0] == WHITESPACE:
                for c in self._catchers.values():
                    c.append(self._peek[1])
                self._peek = next(self._tokenizer)
        except StopIteration:
            self._has_last_value = True

    #def current(self):
    #    """
    #    Returns the current token.
    #    """
    #    return self._next

    def next(self):
        """
        Takes the next token from the stream and returns it.
        """
        if self._has_last_value:
            raise StopIteration
        self._advance()
        return self._next

    def __next__(self):
        return self.next()

    def peek(self):
        """
        Peeks ahead and returns the next token it sees.
        """
        if self._has_last_value:
            raise IndexError
        return self._peek

    def start_catching(self):
        """
        Creates a new buffer that catches the string parts of each token.
        The method returns a handle used to identify this buffer.
        """
        self._catcheri += 1
        self._catchers[self._catcheri] = []
        return self._catcheri

    def stop_catching(self, handle):
        """
        Closes the buffer with the given handle and returns the caught
        input.
        """
        txt = ''.join(self._catchers[handle])
        del self._catchers[handle]
        return txt

    def _tizer(self, stream, check_indenting):

        # All columns are stored to determine the level of indenting
        #  (This may require some extra checks...)
        columns = []

        # Indenting
        dents = 0

        # Parentheses must line up, parentheses can join lines together
        bracket_depth = 0
        bracket_lines = []

        # Lines can be appended with () or \, or """ for meta-values
        append_next_line = False
        in_multi_string = False

        # Block comments can be made by starting a line with """
        in_block_comment = False

        # Comment was found on the current line
        comment = False

        # TEXT also follows the \ rule
        text_buffer = False
        text_start = None

        # Loop through lines
        numb = 0
        ln = 0
        for line in stream:

            # First line number is 1
            numb += 1

            # Not sure this is necessary anymore!
            #if not isinstance(line, bytes):
            #    line = unicodedata.normalize('NFKD', line)

            # Handle multi-line meta-property strings
            if in_multi_string:
                line = line.rstrip()
                p = line.find('"""')
                if p < 0:
                    text_buffer.append(line)
                    continue
                elif p != len(line) - 3:
                    raise ParseError(
                        'Unexpected character(s)',
                        numb, p + 3,
                        'Text found after closing of multi-line string')
                text_buffer.append(line[0:p])
                text = self._post_process_multiline_string(text_buffer)
                yield TEXT, text, text_start[0], text_start[1]
                yield EOL, _sEOL, numb, len(line)
                in_multi_string = False
                text_buffer = False
                continue

            # Handle block comments
            if in_block_comment:
                p = line.find('"""')
                if p < 0:
                    continue
                else:
                    line = line[p + 3:]
                    in_block_comment = False
            elif line[0:3] == '"""':
                p = line.find('"""', 3)
                if p < 0:
                    in_block_comment = True
                    continue
                else:
                    line = line[p + 3:]

            # Ordinary lines, strip comments
            z = line.find('#')
            comment = (z >= 0)
            if comment:
                line = line[0:z]

            # Trim whitespace from end of string
            line = line.rstrip()
            ln = len(line)

            # Skip empty lines
            if ln == 0:
                if append_next_line:
                    # Handle multi-line situation
                    if bracket_depth > 0:
                        # Continue until bracket is closed
                        continue
                    if comment:
                        # Line with comment only doesn't end multi-line
                        # situation!
                        continue
                    if text_buffer:
                        # Multiline TEXT ends on blank line: yield!
                        yield TEXT, ''.join(text_buffer), text_start[0], \
                            text_start[1]
                        text_buffer = False
                    yield EOL, _sEOL, numb, ln
                    append_next_line = False
                continue

            # Index of character in character array
            # First character is 0
            pos = 0

            # Position in line (tab = 8 columns)
            column = 0

            # Initial whitespace is indenting
            countColumns = True

            # Append this line to previous?
            if append_next_line:
                # Skip whitespace
                m = _rWHITE.match(line)
                if m:
                    pos = m.end()
                countColumns = False
                if text_buffer:
                    if line[-1] == '\\':
                        text_buffer.append(line[pos:-1])
                    else:
                        text_buffer.append(line[pos:])
                        yield TEXT, ''.join(text_buffer), text_start[0], \
                            text_start[1]
                        yield EOL, _sEOL, numb, ln
                        append_next_line = False
                        text_buffer = False
                    continue
            append_next_line = False

            # Loop over token matches
            while pos < ln:
                m = _rTOKEN.match(line, pos)
                if m:
                    # Get token, token bounds, first character, move pointer
                    char = line[pos]
                    start, end = m.span()
                    token = line[start:end]
                    size = end - start
                    pos = end

                    # New line? convert whitespace to indent/dedent
                    if countColumns and check_indenting:

                        if char in _sWHITE:
                            for char in token:
                                if char == ' ':
                                    column += 1
                                elif char == '\t':
                                    column = (1 + column // tabsize) * tabsize
                            continue

                        else:

                            if not columns:
                                # First non-whitespace of new file
                                columns.append(column)
                            else:
                                # Differing whitespace?
                                if column > columns[-1]:
                                    # Yield indent token
                                    yield INDENT, line[0:start], numb, start
                                    columns.append(column)
                                    dents += 1
                                elif column < columns[-1]:
                                    # Yield dedent token(s)
                                    while column < columns[-1]:
                                        yield (
                                            DEDENT, line[0:start], numb, start)
                                        dents -= 1
                                        columns.pop()
                                    if column != columns[-1]:
                                        raise ParseError(
                                            'Unexpected indenting level',
                                            numb, start,
                                            'Indenting doesn\'t match with'
                                            ' previous level')
                            # Finished counting columns for this line
                            countColumns = False

                    # Ignore whitespace between tokens
                    if char == ' ' or char == '\t':
                        yield WHITESPACE, token, numb, start
                        continue

                    if size == 1:
                        if char == ':':
                            # Colon? Treat remainder of line as text
                            yield COLON, token, numb, start
                            m = _rWHITE.match(line, end)
                            if m:
                                end = m.end()
                            # Get text, strip whitespace from right
                            text = line[end:].rstrip()
                            if text[0:3] == '"""':
                                # Triple quoted string
                                text = text[3:]
                                end += 3
                                p = text.find('"""')
                                if p < 0:
                                    # Multi-line string
                                    text_buffer = [text]
                                    text_start = numb, end
                                    in_multi_string = True
                                else:
                                    # Single-line string
                                    text = text[:-3]
                                    if p != len(text):
                                        raise ParseError(
                                            'Unexpected character(s)',
                                            numb, end + p + 3,
                                            'Text found after closing of'
                                            ' multi-line string')
                                    text = text.rstrip()
                                    yield TEXT, text, numb, end
                            elif text[-1:] == '\\':
                                # Multi-line string
                                text_buffer = [text[0:-1]]
                                text_start = numb, end
                                append_next_line = True
                            elif text != '':
                                # Single-line string
                                yield TEXT, text, numb, end
                            pos = ln
                            break
                        elif char == '\\':
                            # Append next line to this one
                            if end == ln:
                                append_next_line = True
                                continue
                            else:
                                raise ParseError(
                                    'Illegal character', numb, start,
                                    'Backslash must be last character in line')
                        elif char == '(':
                            bracket_lines.append((numb, start))
                            bracket_depth += 1
                            yield PAREN_OPEN, token, numb, start
                            continue
                        elif char == ')':
                            bracket_depth -= 1
                            if bracket_depth < 0:
                                raise ParseError(
                                    'Parentheses mismatch', numb, start,
                                    'Closing bracket without opening bracket')
                            bracket_lines.pop()
                            yield PAREN_CLOSE, token, numb, start
                            continue
                        else:
                            # Known single character token?
                            index = _SINGLE.find(char)
                            if index >= 0:
                                yield _SINGLE_MAP[index], token, numb, start
                                continue

                    elif size == 2:

                        if token[1] == '=':
                            # Comparison with ?=
                            index = _COMPEQ.find(char)
                            if index >= 0:
                                yield _COMPEQ_MAP[index], token, numb, start
                                continue
                        if token == '//':
                            # Quotient (Integer division)
                            yield QUOTIENT, token, numb, start
                            continue

                    if char == '[':
                        # Segment
                        yield SEGMENT_HEADER, token, numb, start
                        continue

                    if char in _sNUMBERS:
                        # Integer or float
                        m = _rFloat.search(token)
                        if m:
                            yield FLOAT, token, numb, start
                        else:
                            yield INTEGER, token, numb, start
                        continue

                    if char == '.' and size > 1:
                        # Float
                        yield FLOAT, token, numb, start
                        continue

                    if token[-1] == '(':
                        # Function opening
                        # Yield function name, then back up to yield PAREN_OPEN
                        # on next pass
                        yield FUNC_NAME, token[:-1], numb, start
                        pos -= 1
                        continue
                    elif token[-1] == ':':
                        # Meta property name
                        # Yield name, then back up to yield COLON on next pass
                        yield META_NAME, token[:-1].strip(), numb, start
                        pos -= 1
                        continue
                    elif token in KEYWORD_MAP:
                        # Reserved keyword (and, or etc.)
                        yield KEYWORD_MAP[token], token, numb, start
                        continue
                    else:
                        # Variable name
                        yield NAME, token, numb, start
                        continue

                else:
                    # No token matched

                    token = line[pos:]
                    raise ParseError(
                        'Unknown or invalid token', numb, pos,
                        'Unrecognized token: ' + token)

            # Join lines with open parentheses
            if bracket_depth > 0:
                append_next_line = True
            else:
                # Yield end-of-line
                if not (append_next_line or in_multi_string):
                    yield EOL, _sEOL, numb, ln

        # End of file, test parentheses mismatch
        if bracket_depth > 0:
            numb, pos = bracket_lines.pop()
            raise ParseError(
                'Parentheses mismatch', numb, pos,
                'Parentheses opened but never closed')
        if in_multi_string:
            raise ParseError(
                'Unclosed multi-line string', text_start[0], text_start[1] - 3,
                'Multi-line string opened but never closed')

        # De-dent at end of file
        if check_indenting:
            while dents > 0:
                dents -= 1
                yield DEDENT, '', numb, ln
        yield EOF, '', numb, ln

    def _post_process_multiline_string(self, text_buffer):
        """
        Post-processes a multi line meta value, removing unnecessary
        whitespace.

        Any value given using the triple-quote syntax must be pre-processed.
        This function finds the lowest indentation level used in the string's
        declaration and trims the corresponding whitespace from each line. The
        result is then returned, with the line breaks and everything else left
        intact.
        """
        # Find lowest whitespace level, trim from left
        ind = 'first'
        for line in text_buffer:
            if ind == 'first':
                # Skip first line
                ind = None
            elif line.strip() == '':
                # Skip blank lines
                continue
            else:
                # Find whitespace level, update global
                white = _rWHITE.match(line)
                white = line[0:white.end()]
                pos = 0
                column = 0
                for char in white:
                    if char == ' ':
                        column += 1
                    elif char == '\t':
                        column = (1 + column // tabsize) * tabsize
                if ind is None or column < ind:
                    ind = column

        # No indentation? Then return as is
        if ind == 0 or ind is None or ind == 'first':
            return '\n'.join(text_buffer)

        # Strip initial ``ind`` characters from lines
        text = []
        first = True
        for line in text_buffer:
            if first:
                text.append(line.lstrip())
                first = False
            elif line.strip() == '':
                text.append('')
            else:
                pos = 0
                column = 0
                while column < ind:
                    if line[pos] == ' ':
                        column += 1
                    elif line[pos] == '\t':
                        column = (1 + column // tabsize) * tabsize
                    else:   # pragma: no cover
                        raise Exception(
                            'Unexpected character in multi-line string\'s'
                            ' whitespace: "' + line[pos] + '"')
                    pos += 1
                text.append(line[pos:])
        return ('\n'.join(text)).strip()


def parse_proto_expression(stream, info=None, rbp=0):
    """
    Parses an expression, requires a :class:`TokenStream` ``stream``, a
    :class:`ParseInfo` object ``info`` and an integer ``rbp`` representing the
    current right-binding-power.

    The function will return only when all tokens are exhausted or a lower
    binding power is encountered.

    Returns a "proto-expression"; a tuple ``(Type, Arguments, Tokens)``. Here
    ``Type`` is the expression type (either a :class:`myokit.Expression` or the
    name of a user function). ``Arguments`` is a sequence of constructor
    arguments and ``Tokens`` is a sequence of tokens.
    """
    if info is None:
        info = ParseInfo()

    # Parse first token using null denomination
    code, text, line, char = stream.peek()
    try:
        parser = nud_parsers[code]
    except KeyError:
        unexpected_token(stream.peek(), 'expression')
    expr = parser.parse(stream, info)

    # Parse follow up tokens
    code, text, line, char = stream.peek()
    while True:
        try:
            parser = led_parsers[code]
        except KeyError:
            break
        if parser._rbp <= rbp:
            break
        expr = parser.parse(expr, stream, info)
        code, text, line, char = stream.peek()
    return expr


def parse_expression_stream(stream, context=None):
    """
    Parses an expression from a stream. A :class:`myokit.Variable` object can
    be given as ``context`` to resolve any references against.
    """
    return convert_proto_expression(parse_proto_expression(stream), context)


def parse_expression_string(string, context=None):
    """
    Parses string data into a :class:`myokit.Expression`.

    A :class:`myokit.Variable` object can be given as ``context`` to resolve
    any references against.
    """
    # Create fino from context variable, if given
    # (Required when parsing user functions given as strings)
    info = None
    if context is not None:
        info = ParseInfo()
        if isinstance(context, myokit.Model):
            info.model = context
        else:
            info.model = context.model()

    # Tokenise and parse string to proto expression
    s = Tokenizer(string)
    e = parse_proto_expression(s, info=info)

    # Check for eol, eof, then nothing else
    expect(next(s), EOL)
    expect(next(s), EOF)

    # Convert proto expression and return
    return convert_proto_expression(e, context, info)


#def parse_number_string(string):
#    """
#    Parses string data into a :class:`myokit.Number`.
#    """
#    s = Tokenizer(string)
#    p = NumberParser()
#    e = p.parse(s, ParseInfo())
#    expect(next(s), EOL)
#    expect(next(s), EOF)
#    try:
#        next(s)
#        raise ParseError(
#            'Unused tokens', 0, 0, 'Expecting a string containing only a'
#            ' single number.')
#    except StopIteration:
#        return convert_proto_expression(e)


def convert_proto_expression(e, context=None, info=None):
    """
    Converts a proto-expression to a :class:`myokit.Expression`. Requires a
    :class:`myokit.Variable` object ``context`` if any references are to be
    resolved. If a :class:`ParseInfo` object ``info`` is passed in it will be
    used to register tokens.
    """
    def convert(x):
        # Unpack proto expression
        element, ops, tokens = x

        # Create expression
        if element == myokit.Number:
            # Handle Numbers
            e = myokit.Number(*ops)

        elif element == myokit.Name:
            # Handle Names, resolve references
            if context is None:
                # No variable given to determine scope and resolve references
                # with, return a string-based name for function templates (used
                # for user functions) and debugging.
                return myokit.Name(*ops)

            # Resolve reference
            # Note: I tried caching resolved references per variable, but it
            # didn't speed-up or slow down execution time.
            try:
                e = myokit.Name(context._resolve(ops[0]))
            except myokit.UnresolvedReferenceError as e:
                a, b = tokens[0][2:4] if tokens else (0, 0)
                raise ParseError('Unresolved reference', a, b, str(e), cause=e)

        elif isinstance(element, myokit.UserFunction):
            # Handle user function
            # Get mapping of function argument names to input values
            args = {}
            for k, p in enumerate(element.arguments()):
                args[p] = convert_proto_expression(ops[k], context, info)
            e = element.convert(args)

        else:
            # Handle other types
            ops = [convert(op) for op in ops]
            try:
                e = element(*ops)
            except myokit.IntegrityError as e:
                line, char = tokens[0][2:4] if tokens else (0, 0)
                raise ParseError('Syntax error', line, char, str(e), cause=e)

        # Register tokens
        if info:
            for token in tokens:
                if token:
                    reg_token(info, token, e)
        return e

    return convert(e)


def format_parse_error(ex, source=None):
    """
    Turns a ParseError ``ex`` into a detailed error message.

    If a filename or input stream is passed in as ``source``, this will
    be used to show the line on which an error occurred. If a stream is passed,
    it should be rewinded to the same point the original parsing started.
    """
    out = [ex.name]
    if ex.desc is not None:
        out.append('  ' + ex.desc)
    out.append('On line ' + str(ex.line) + ' character ' + str(ex.char))
    line = None

    if ex.line > 0 and source is not None:
        if isinstance(source, str) and os.path.isfile(source):
            # Re-open file, find line
            with open(source, 'r') as f:
                for i in range(0, ex.line):
                    line = next(f)
            line = line.rstrip()
        else:
            i = 0
            for line in source:
                i += 1
                if i == ex.line:
                    break
            if i != ex.line:
                line = None

    if line is not None and len(line) > ex.char:
        # Skip initial whitespace
        pos = 0
        _sWHITE = ' \t'
        for char in line[0:ex.char]:
            if char not in _sWHITE:
                break
            pos += 1

        # Add line
        line = line[pos:].expandtabs(tabsize)
        char = ex.char - pos
        n = len(line)
        if n > 56:
            # Trim if too long
            p2 = min(n, ex.char + 30)
            p1 = p2 - 56
            if p1 < 0:
                p2 -= p1
                p1 = 0
            line = line[p1:p2]
            char = char - p1
            if p1 > 0:
                line = '..' + line
                char += 2
            if p2 < n:
                line += '..'
        out.append('  ' + line)

        # Add error indication
        out.append(' ' * (2 + char) + '^')

    return '\n'.join(out)


class NudParser:
    """
    Expression parser for nud operators.

    An expression parser to use when parsing a token as the first element
    of an expression. The term nud stands for 'null denotation' as it has no
    previous expression bound to it.
    """
    def __init__(self, element=None):
        self.element = element
        self._rbp = element._rbp if element else 0

    def parse(self, stream, info):
        raise NotImplementedError


class NumberParser(NudParser):
    """
    Parser for numeric literals.
    """
    def parse(self, stream, info):
        token = next(stream)
        unit = None
        if stream.peek()[0] == BRACKET_OPEN:
            unit = parse_bracketed_unit(stream)
        return (myokit.Number, (token[1], unit), (token,))


class NameParser(NudParser):
    """
    Parser for names.
    """
    def parse(self, stream, info):
        t1 = next(stream)
        t2 = t3 = t4 = None
        name = t1[1]
        if stream.peek()[0] == DOT:
            t2 = next(stream)
            t3 = expect(next(stream), NAME)
            name += '.' + t3[1]
        return (myokit.Name, (name, ), (t1, t2, t3, t4))


class PrefixParser(NudParser):
    """
    Parser for prefix (single operand) operators.
    """
    def parse(self, stream, info):
        token = next(stream)
        arg = parse_proto_expression(stream, info, self._rbp)
        return (self.element, (arg, ), (token,))


class GroupingParser(NudParser):
    """
    Parser for grouping IE parentheses; as in 5 * (2 + 3).
    """
    def parse(self, stream, info):
        next(stream)
        expr = parse_proto_expression(stream, info, self._rbp)
        expect(next(stream), PAREN_CLOSE)
        return expr


class FunctionParser(NudParser):
    """
    Parser for function calls.
    """
    def __init__(self):
        super().__init__()
        self._rbp = myokit.Function._rbp

    def parse(self, stream, info):
        name = next(stream)
        ops = []
        token = next(stream)
        while token[0] != PAREN_CLOSE:
            ops.append(parse_proto_expression(stream, info))
            token = expect(next(stream), [COMMA, PAREN_CLOSE])

        # Predefined function
        if name[1] in functions:
            func = functions[name[1]]
            if func._nargs is not None:
                # Allow number-of-arguments check to be bypassed
                if not len(ops) in func._nargs:
                    raise ParseError(
                        'Syntax error', name[2], name[3],
                        'Wrong number of arguments for function '
                        + str(func._fname) + '()')
            return (func, ops, (name,))

        # User-defined function
        func = None
        if info.model is not None:
            try:
                func = info.model.get_function(name[1], len(ops))
            except KeyError:
                pass
        if func is None:
            raise ParseError(
                'Unknown function', name[2], name[3], 'A function '
                + name[1] + '() with ' + str(len(ops))
                + ' argument(s) could not be found.')

        # Found function, return template, arguments and tokens. "func" is
        # now a (template) Expression object.
        return (func, ops, (name,))


class LedParser:
    """
    Expression parser for led operators.

    An expression parser to use when parsing a token with an existing
    expression. The term led stands for 'left denotation' as it is bound to
    an existing expression on its left.

    Arguments:

    ``element``
        The model element this parser will create.
    ``rbp``
        This parser's right binding power. A ``LedParser`` will absorb tokens
        on its right as long as it has more binding power than the token
        on its left.
    """
    def __init__(self, element):
        self.element = element
        self._rbp = element._rbp if element else 0

    def parse(self, left, stream, info):
        raise NotImplementedError


class InfixParser(LedParser):
    """
    Parser for infix operators.
    """
    def parse(self, left, stream, info):
        token = next(stream)
        right = parse_proto_expression(stream, info, self._rbp)
        return (self.element, (left, right), (token,))


# Null denotation parsers
nud_parsers = {}
nud_parsers[NAME] = NameParser()
nud_parsers[INTEGER] = NumberParser()
nud_parsers[FLOAT] = NumberParser()
nud_parsers[PLUS] = PrefixParser(myokit.PrefixPlus)
nud_parsers[MINUS] = PrefixParser(myokit.PrefixMinus)
nud_parsers[NOT] = PrefixParser(myokit.Not)
nud_parsers[PAREN_OPEN] = GroupingParser()
nud_parsers[FUNC_NAME] = FunctionParser()
# Left denomination parsers
led_parsers = {}
led_parsers[PLUS] = InfixParser(myokit.Plus)
led_parsers[MINUS] = InfixParser(myokit.Minus)
led_parsers[STAR] = InfixParser(myokit.Multiply)
led_parsers[SLASH] = InfixParser(myokit.Divide)
led_parsers[REMAINDER] = InfixParser(myokit.Remainder)
led_parsers[QUOTIENT] = InfixParser(myokit.Quotient)
led_parsers[POWER] = InfixParser(myokit.Power)
led_parsers[EQEQUAL] = InfixParser(myokit.Equal)
led_parsers[NOTEQUAL] = InfixParser(myokit.NotEqual)
led_parsers[MORE] = InfixParser(myokit.More)
led_parsers[LESS] = InfixParser(myokit.Less)
led_parsers[MOREEQUAL] = InfixParser(myokit.MoreEqual)
led_parsers[LESSEQUAL] = InfixParser(myokit.LessEqual)
led_parsers[AND] = InfixParser(myokit.And)
led_parsers[OR] = InfixParser(myokit.Or)
# Mathematical functions
functions = {}
functions['sqrt'] = myokit.Sqrt
functions['sin'] = myokit.Sin
functions['cos'] = myokit.Cos
functions['tan'] = myokit.Tan
functions['asin'] = myokit.ASin
functions['acos'] = myokit.ACos
functions['atan'] = myokit.ATan
functions['exp'] = myokit.Exp
functions['log'] = myokit.Log
functions['log10'] = myokit.Log10
functions['floor'] = myokit.Floor
functions['ceil'] = myokit.Ceil
functions['abs'] = myokit.Abs
# Further functions
functions['dot'] = myokit.Derivative
functions['if'] = myokit.If
functions['piecewise'] = myokit.Piecewise
# TODO: 'partial' and 'init' are also function names used in output, but are
# not currently handled by the parser. Not sure if they should be!



#!/usr/bin/env python3
#
# Tests the OpenCL simulation classes
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import OpenCL_DOUBLE_PRECISION, DIR_DATA

# Show simulation output
debug = False


@unittest.skipIf(
    not OpenCL_DOUBLE_PRECISION,
    'OpenCL double precision extension not supported on selected device.')
class SimulationOpenCL1dTest(unittest.TestCase):
    """
    Tests the OpenCL simulation against Sim1d.
    """
    def test_simple_run(self):
        # Compare the SimulationOpenCL output with Simulation1d output

        # Set up simulations
        m = myokit.load_model(os.path.join(DIR_DATA, 'beeler-1977-model.mmt'))
        m.get('membrane.i_stim.amplitude').set_rhs(-80)
        p = myokit.pacing.blocktrain(period=1000, duration=2, offset=3)
        n = 10
        sa = myokit.Simulation1d(m, p, ncells=n)
        sb = myokit.SimulationOpenCL(
            m, p, ncells=n, precision=myokit.DOUBLE_PRECISION)

        dt = 0.005
        sa.set_step_size(dt)
        sb.set_step_size(dt)
        sa.set_paced_cells(2)
        sb.set_paced_cells(2)

        tmax = 15
        logvars = [
            'engine.time',
            'membrane.V',
            'engine.pace',
            'isi.Isi',
            'membrane.i_diff',
        ]
        da = sa.run(tmax, log=logvars, log_interval=0.5).npview()
        db = sb.run(tmax, log=logvars, log_interval=0.5).npview()

        # Show results
        if debug:
            import matplotlib.pyplot as plt
            fig = plt.figure(figsize=(12, 9))
            fig.subplots_adjust(0.08, 0.08, 0.98, 0.98, 0.5, 0.3)
            ax = fig.add_subplot(2, 4, 1)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('V')
            ax.plot(da.time(), da['membrane.V', 0], '-', lw=2, alpha=0.5)
            ax.plot(da.time(), da['membrane.V', 9], '-', lw=2, alpha=0.5)
            ax.plot(db.time(), db['membrane.V', 0], '--')
            ax.plot(db.time(), db['membrane.V', 9], '--')

            ax = fig.add_subplot(2, 4, 2)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('Idiff')
            ax.plot(da.time(), da['membrane.i_diff', 0], '-', lw=2, alpha=0.5)
            ax.plot(da.time(), da['membrane.i_diff', 9], '-', lw=2, alpha=0.5)
            ax.plot(db.time(), db['membrane.i_diff', 0], '--')
            ax.plot(db.time(), db['membrane.i_diff', 9], '--')

            ax = fig.add_subplot(2, 4, 3)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('Isi')
            ax.plot(da.time(), da['isi.Isi', 0], '-', lw=2, alpha=0.5)
            ax.plot(da.time(), da['isi.Isi', 9], '-', lw=2, alpha=0.5)
            ax.plot(db.time(), db['isi.Isi', 0], '--')
            ax.plot(db.time(), db['isi.Isi', 9], '--')

            ax = fig.add_subplot(2, 4, 4)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('pace')
            ax.plot(da.time(), da['engine.pace'], '-', lw=2, alpha=0.5)
            ax.plot(db.time(), db['engine.pace'], '--')

            ax = fig.add_subplot(2, 4, 5)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('V error')
            ax.plot(da.time(), da['membrane.V', 0] - db['membrane.V', 0])
            ax.plot(da.time(), da['membrane.V', 9] - db['membrane.V', 9])

            ax = fig.add_subplot(2, 4, 6)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('Idiff error')
            ax.plot(da.time(),
                    da['membrane.i_diff', 0] - db['membrane.i_diff', 0])
            ax.plot(da.time(),
                    da['membrane.i_diff', 9] - db['membrane.i_diff', 9])

            ax = fig.add_subplot(2, 4, 7)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('Isi error')
            ax.plot(da.time(), da['isi.Isi', 0] - db['isi.Isi', 0])
            ax.plot(da.time(), da['isi.Isi', 9] - db['isi.Isi', 9])

            ax = fig.add_subplot(2, 4, 8)
            ax.set_xlabel('Time (ms)')
            ax.set_ylabel('pace error')
            ax.plot(da.time(), da['engine.pace'] - db['engine.pace'])
            plt.show()

        e = np.abs(da.time() - db.time())
        self.assertLess(np.max(e), 1e-17)
        e = np.abs(da['engine.pace'] - db['engine.pace'])
        self.assertLess(np.max(e), 1e-17)

        e = np.abs(da['membrane.V', 0] - db['membrane.V', 0])
        self.assertLess(np.max(e), 1e-13)
        e = np.abs(da['membrane.V', 9] - db['membrane.V', 9])
        self.assertLess(np.max(e), 1e-13)

        e = np.abs(da['membrane.i_diff', 0] - db['membrane.i_diff', 0])
        self.assertLess(np.max(e), 1e-13)
        e = np.abs(da['membrane.i_diff', 9] - db['membrane.i_diff', 9])
        self.assertLess(np.max(e), 1e-13)

        e = np.abs(da['isi.Isi', 0] - db['isi.Isi', 0])
        self.assertLess(np.max(e), 1e-13)
        e = np.abs(da['isi.Isi', 9] - db['isi.Isi', 9])
        self.assertLess(np.max(e), 1e-13)


if __name__ == '__main__':
    import sys
    if '-v' in sys.argv:
        print('Running in debug/verbose mode')
        debug = True
    else:
        print('Add -v for more debug output')
    unittest.main()
